#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Set of functions to interface with the neuromorpho.org database of
neurons.

See http://neuromorpho.org/apiReference.html for documentation.
"""

import requests
import urllib

import pandas as pd

from tqdm import tqdm
from typing import List, Dict, Union

from ..core import TreeNeuron, NeuronList
from ..io.swc_io import from_swc
from .. import utils


baseurl = 'http://neuromorpho.org'


def list_neurons(**filters) -> pd.DataFrame:
    """ List neurons filtered by given criteria
    """

    if not filters:
        answer = ""
        while answer not in ["y", "n"]:
            answer = input("No filters will list all neurons. Continue? [Y/N] ").lower()

        if answer != 'y':
            return  # type: ignore

    # Turn lists into strings
    filters = {k: ','.join(v) if isinstance(v, list) else v for k, v in filters.items()}

    # Turn filters into str
    fstring = '&'.join([f'{k}:{v}' for k, v in filters.items()])

    url = utils.make_url(baseurl, 'neuron', 'select', q=str(filters))


def get_neuron_info(x: Union[str, int]) -> pd.Series:
    """ Fetch neuron info by ID or by name.


    Parameters
    ----------
    x :         int | str
                Integer is intepreted as ID, string as neuron name. Will try
                to convert strings to integers first.

    Examples
    --------
    >>> import navis.interfaces.neuromorpho as nm
    >>> # Get info by ID
    >>> info = nm.get_neuron_info(1)
    >>> # Get info by Name
    >>> info = nm.get_neuron_info('cnic_001')
    """

    try:
        x = int(x)
    except BaseException:
        pass

    if isinstance(x, str):
        url = utils.make_url(baseurl, 'api', 'neuron', 'name', x)
    elif isinstance(x, int):
        url = utils.make_url(baseurl, 'api', 'neuron', 'id', str(x))
    else:
        raise TypeError(f'Expected string or int, got {type(x)}')

    resp = requests.get(url)

    resp.raise_for_status()

    return pd.Series(resp.json())


def get_neuron(x: Union[str, int, Dict[str, str]], **kwargs) -> TreeNeuron:
    """ Fetch neuron by ID or by name.


    Parameters
    ----------
    x :         int | str | dict
                Integer is intepreted as ID, string as neuron name. Dictionary
                must contain 'archive' (e.g. "Wearne_Hof") and 'neuron_name'
                (e.g. "cnic_001").
    **kwargs
                Keyword arguments passed on to :func:`navis.from_swc`.

    Returns
    -------
    TreeNeuron

    Examples
    --------
    >>> import navis.interfaces.neuromorpho as nm
    >>> # Get a neuron by its ID
    >>> n = nm.get_neuron(1)
    >>> n
    type            TreeNeuron
    name                   SWC
    n_nodes               1274
    n_connectors             0
    n_branches              46
    n_leafs                 54
    cable_length       4792.21
    soma                  None
    """

    if not isinstance(x, (pd.Series, dict)):
        info = get_neuron_info(x)
    else:
        info = x  # type: ignore

    archive: str = info['archive']
    name: str = info['neuron_name']

    url = utils.make_url(baseurl, 'dableFiles', archive.lower(), 'CNG version', name + '.CNG.swc')

    return from_swc(url, **kwargs)


def get_neuron_fields() -> Dict[str, List[str]]:
    """ Returns a list of neuron fields avaialble.

    Examples
    --------
    >>> import navis.interfaces.neuromorpho as nm
    >>> fields = nm.get_neuron_fields()

    """

    url = utils.make_url(baseurl, 'api', 'neuron', 'fields')
    resp = requests.get(url)

    resp.raise_for_status()

    return resp.json().get('Neuron Fields')


def get_available_field_values(field: str) -> List[str]:
    """ Returns a list of values present in the repository for the neuron
    field requested.

    Parameters
    ----------
    field :     str
                Field to search for

    Examples
    --------
    >>> import navis.interfaces.neuromorpho as nm
    >>> # Get availalbe values for "species" field
    >>> species = nm.get_available_field_values('species')
    """

    data: List[str] = []
    page = 1

    with tqdm(total=1, desc='Fetching') as pbar:
        while True:
            url = utils.make_url(baseurl, 'api', 'neuron', 'fields', field, page=page)

            resp = requests.get(url)

            resp.raise_for_status()

            content = resp.json()

            data += content['fields']

            if page == content['page']['totalPages']:
                break

            pbar.total = content['page']['totalPages']
            pbar.update(1)

            page += 1

    return data



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

from .converters import (network2nx, network2igraph, neuron2igraph, nx2neuron,
                         neuron2nx, neuron2KDTree, neuron2dps)
from .graph_utils import (classify_nodes, cut_neuron, longest_neurite,
                          split_into_fragments, reroot_neuron, distal_to,
                          dist_between, find_main_branchpoint,
                          generate_list_of_childs, geodesic_matrix,
                          subset_neuron, node_label_sorting, _break_segments,
                          _generate_segments, segment_length,
                          _connected_components)
from .clinic import (health_check)



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Module contains functions to plot neurons in 1D.
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

import numpy as np

from typing import Optional, Union, Dict, Tuple, Any

from .. import core, config, graph

logger = config.logger

colortype = Union[str,
                  Tuple[float, float, float],
                  Tuple[float, float, float, float]]


def plot1d(x: 'core.NeuronObject',
           ax: Optional['matplotlib.ax'] = None,
           color: Optional[Union['str',
                                 colortype,
                                 Dict[Any, colortype],
                                 ]
                           ] = None,
           **kwargs) -> mpl.axes.Axes:
    """ Plot neuron topology in 1D according to Cuntz et al. (2010).

    This function breaks a neurons into segments between branch points.
    See Cuntz et al., PLoS Computational Biology (2010) for detailed
    explanation. For very complex neurons, this neuron "barcode" can get
    fairly complicated - make sure to zoom in.

    Parameters
    ----------
    x :         TreeNeuron | NeuronList
                Neuron(s) to plot.
    ax :        matplotlib.ax, optional
    cmap :      tuple | dict
                Color. If dict must map neuron UUID to color.
    **kwargs
                Will be passed to ``matplotlib.patches.Rectangle``.

    Returns
    -------
    matplotlib.ax

    Examples
    --------
    >>> import navis
    >>> import matplotlib.pyplot as plt
    >>> n = navis.example_neuron(2)
    >>> ax = navis.plot1d(n)
    >>> plt.show()

    """

    if isinstance(x, core.NeuronList):
        pass
    elif isinstance(x, core.TreeNeuron):
        x = core.NeuronList(x)
    else:
        raise TypeError(f'Unable to work with data of type "{type(x)}"')

    if isinstance(color, type(None)):
        color = (0.56, 0.86, 0.34)

    if not ax:
        fig, ax = plt.subplots(figsize=(8, len(x) / 3))

    # Add some default parameters for the plotting to kwargs
    kwargs.update({'lw': kwargs.get('lw', .1),
                   'ec': kwargs.get('ec', (1, 1, 1)),
                   })

    max_x = []
    for ix, n in enumerate(config.tqdm(x, desc='Processing',
                                       disable=config.pbar_hide,
                                       leave=config.pbar_leave)):
        if isinstance(color, dict):
            this_c = color[n.id]
        else:
            this_c = color

        # Get topological sort (root -> terminals)
        topology = graph.node_label_sorting(n)

        # Get terminals and branch points
        bp = n.nodes[n.nodes.type == 'branch'].node_id.values
        term = n.nodes[n.nodes.type == 'end'].node_id.values

        # Order this neuron's segments by topology
        breaks = [topology[0]] + \
            [n for i, n in enumerate(topology) if n in bp or n in term]
        segs = [([s for s in n.small_segments if s[0] == end][0][-1], end)
                for end in breaks[1:]]

        # Now get distances for each segment
        if 'nodes_geodesic_distance_matrix' in n.__dict__:
            # If available, use geodesic distance matrix
            dist_mat = n.nodes_geodesic_distance_matrix
        else:
            # If not, compute matrix for subset of nodes
            dist_mat = graph.geodesic_matrix(n, tn_ids=breaks, directed=False)

        dist = np.array([dist_mat.loc[s[0], s[1]] for s in segs]) / 1000
        max_x.append(sum(dist))

        # Plot
        curr_dist = 0
        for k, d in enumerate(dist):
            if segs[k][1] in term:
                c = tuple(np.array(this_c) / 2)
            else:
                c = color

            p = mpatches.Rectangle((curr_dist, ix), d, 1, fc=c, **kwargs)
            ax.add_patch(p)
            curr_dist += d

    ax.set_xlim(0, max(max_x))
    ax.set_ylim(0, len(x))

    ax.set_yticks(np.array(range(0, len(x))) + .5)
    ax.set_yticklabels(x.name)

    ax.set_xlabel('distance [um]')

    ax.set_frame_on(False)

    try:
        plt.tight_layout()
    except BaseException:
        pass

    return ax



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import json

import pandas as pd

from .. import config, core

# Set up logging
logger = config.logger


def neuron2json(x: 'core.NeuronObject', **kwargs) -> str:
    """ Generate JSON formatted ``str`` respresentation of TreeNeuron/List.

    Nodes and connectors are serialised using pandas' ``to_json()``. Most
    other items in the neuron's __dict__ are serialised using
    ``json.dumps()``. Properties not serialised: `.graph`, `.igraph`.

    Parameters
    ----------
    x :         TreeNeuron | NeuronList
    **kwargs
                Parameters passed to ``json.dumps()`` and
                ``pandas.DataFrame.to_json()``.

    Returns
    -------
    str

    See Also
    --------
    :func:`~navis.json2neuron`
                Read json back into navis neurons.

    """

    if not isinstance(x, (core.TreeNeuron, core.NeuronList)):
        raise TypeError('Unable to convert data of type "{0}"'.format(type(x)))

    if isinstance(x, core.TreeNeuron):
        x = core.NeuronList([x])

    data = []
    for n in x:
        this_data = {'skeleton_id': n.skeleton_id}

        if 'nodes' in n.__dict__:
            this_data['nodes'] = n.nodes.to_json()

        if 'connectors' in n.__dict__:
            this_data['connectors'] = n.connectors.to_json(**kwargs)

        for k in n.__dict__:
            if k in ['nodes', 'connectors', 'graph', 'igraph', '_remote_instance',
                     'segments', 'small_segments', 'nodes_geodesic_distance_matrix',
                     'dps', 'simple']:
                continue
            try:
                this_data[k] = n.__dict__[k]
            except BaseException:
                logger.error('Lost attribute "{0}"'.format(k))

        data.append(this_data)

    return json.dumps(data, **kwargs)


def json2neuron(s: str, **kwargs) -> 'core.NeuronList':
    """ Load neuron from JSON string.

    Parameters
    ----------
    s :         str
                JSON-formatted string.
    **kwargs
                Parameters passed to ``json.loads()`` and
                ``pandas.DataFrame.read_json()``.

    Returns
    -------
    :class:`~navis.NeuronList`

    See Also
    --------
    :func:`~navis.neuron2json`
                Turn neuron into json.

    """

    if not isinstance(s, str):
        raise TypeError('Need str, got "{0}"'.format(type(s)))

    data = json.loads(s, **kwargs)

    nl = core.NeuronList([])

    for n in data:
        # Make sure we have all we need
        REQUIRED = ['skeleton_id']

        missing = [p for p in REQUIRED if p not in n]

        if missing:
            raise ValueError('Missing data: {0}'.format(','.join(missing)))

        cn = core.TreeNeuron(int(n['skeleton_id']))

        if 'nodes' in n:
            cn.nodes = pd.read_json(n['nodes'])
            cn.connectors = pd.read_json(n['connectors'])

        for key in n:
            if key in ['skeleton_id', 'nodes', 'connectors']:
                continue
            setattr(cn, key, n[key])

        nl += cn

    return nl



from .volumes import Volume
from .dotprops import Dotprops
from .neurons import Neuron, TreeNeuron
from .neuronlist import NeuronList

from typing import Union

NeuronObject = Union[NeuronList, TreeNeuron]



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import logging
import matplotlib.cm as mcl
import numpy as np
import pint

logger = logging.getLogger('navis')
logger.setLevel(logging.INFO)
if len(logger.handlers) == 0:
    sh = logging.StreamHandler()
    sh.setLevel(logging.DEBUG)
    # Create formatter and add it to the handlers
    formatter = logging.Formatter(
        '%(levelname)-5s : %(message)s (%(name)s)')
    sh.setFormatter(formatter)
    logger.addHandler(sh)

# Default settings for progress bars
pbar_hide = False
pbar_leave = False

# Default settings for caching
warn_caching = True

# Default setting for igraph:
#   If True, will use iGraph if possible
#   If False, will ignore iGraph even if present
# Primarily used for debugging
use_igraph = True

# Default color for neurons
default_color = (.95, .65, .04)

# Unit registry
ureg = pint.UnitRegistry()

# Default connector color palette
default_connector_colors = {
    0: {'name': 'Presynapses',
        'color': (1, 0, 0)},
    1: {'name': 'Postsynapses',
        'color': (0, 0, 1)},
    2: {'name': 'Gap junctions',
        'color': (0, 1, 0)},
    'display': 'lines'  # can also be 'circle'
                        }
# Set some synonyms
default_connector_colors['pre'] = default_connector_colors['Pre'] = default_connector_colors[0]
default_connector_colors['post'] = default_connector_colors['Post'] = default_connector_colors[1]
default_connector_colors['gap'] = default_connector_colors['Gap'] = default_connector_colors[0]
default_connector_colors['gap_junction'] = default_connector_colors['Gap_junction'] = default_connector_colors[0]
default_connector_colors['gap_junctions'] = default_connector_colors['Gap_junctions'] = default_connector_colors[0]


def _type_of_script():
    """Returns context in which navis is run. """
    try:
        ipy_str = str(type(get_ipython()))
        if 'zmqshell' in ipy_str:
            return 'jupyter'
        if 'terminal' in ipy_str:
            return 'ipython'
    except BaseException:
        return 'terminal'


def is_jupyter():
    """Test if navis is run in a Jupyter notebook."""
    return _type_of_script() == 'jupyter'

# Here, we import tqdm and determine whether we use classic notebook tbars
from tqdm import tqdm_notebook, tnrange
from tqdm import tqdm as tqdm_classic
from tqdm import trange as trange_classic

# Keep this because `tqdm_notebook` is only a wrapper (type "function")
tqdm_class = tqdm_classic

if is_jupyter():
    from tqdm import tqdm_notebook, tnrange
    tqdm = tqdm_notebook
    trange = tnrange
else:
    tqdm = tqdm_classic
    trange = trange_classic



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import colorsys
from functools import wraps
import platform
import uuid
import warnings

import matplotlib.colors as mcl
import numpy as np
import pandas as pd
import png
import scipy.spatial
import seaborn as sns

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import vispy as vp

from ... import utils, config

__all__ = ['Viewer']

logger = config.logger


class Browser:
    """ Vispy browser.

    Parameters
    ----------
    n_rows/n_cols : int, optional
                    Number of rows and columns in the browser.

    """

    def __init__(self, n_rows=2, n_cols=3, link=False, **kwargs):
        # Update some defaults as necessary
        defaults = dict(keys=None,
                        show=True,
                        bgcolor='white')
        defaults.update(kwargs)

        # Generate canvas
        self.canvas = vp.scene.SceneCanvas(**defaults)

        # Add grid
        self.grid = self.canvas.central_widget.add_grid()

        # Add 3d widgets
        self.viewers = []
        for i in range(n_rows):
            for k in range(n_cols):
                v = self.grid.add_view(row=i, col=k)
                v.camera = 'turntable'
                v.border_color = (.5, .5, .5, 1)
                self.viewers.append(v)

        if link:
            for v in self.viewers[1:]:
                self.viewers[0].camera.link(v.camera)

    def add_to_all(self, x, center=True, clear=False, **kwargs):
        """ Add visuals to all viewers. """
        for v in range(len(self.viewers)):
            self.add(x, viewer=v, center=center, clear=clear, **kwargs)

    def add_and_divide(self, x, center=True, clear=False, **kwargs):
        """ Divide up visuals onto available viewers. """
        data = utils._parse_objects(x)

        v = 0
        for d in data:
            if isinstance(d, pd.DataFrame):
                d = d.itertuples()
            for to_add in d:
                self.add(to_add, viewer=v, center=center,
                         clear=clear, **kwargs)
                v += 1

                if v == len(self.viewers):
                    v = 0

    def add(self, x, viewer=1, center=True, clear=False, **kwargs):
        """ Add objects to canvas.

        Parameters
        ----------
        x :         skeleton IDs | Neuron/List | Dotprops | Volumes | Points | vispy visuals
                    Object(s) to add to the canvas.
        viewer :    int | slice, optional
                    Index of the viewer to add object to.
        center :    bool, optional
                    If True, re-center camera to all objects on canvas.
        clear :     bool, optional
                    If True, clear canvas before adding new objects.
        **kwargs
                    Keyword arguments passed when generating visuals. See
                    :func:`~navis.plot3d` for options.

        Returns
        -------
        None
        """

        (skids, skdata, dotprops, volumes,
         points, visuals) = utils._parse_objects(x)

        colors = kwargs.get('color',
                            kwargs.get('c',
                                       kwargs.get('colors', None)))
        # Parse colors for neurons and dotprops
        neuron_cmap, skdata_cmap = plotting._prepare_colormap(colors,
                                                              skdata, dotprops,
                                                              use_neuron_color=kwargs.get('use_neuron_color', False))
        kwargs['color'] = neuron_cmap + skdata_cmap

        if skids:
            visuals += plotting._neuron2vispy(fetch.get_neurons(skids),
                                              **kwargs)
        if skdata:
            visuals += plotting._neuron2vispy(skdata, **kwargs)
        if not dotprops.empty:
            visuals += plotting._dp2vispy(dotprops, **kwargs)
        if volumes:
            visuals += plotting._volume2vispy(volumes, **kwargs)
        if points:
            visuals += plotting._points2vispy(points,
                                              **kwargs.get('scatter_kws', {}))

        if not visuals:
            raise ValueError('No visuals generated.')

        if clear:
            self.clear()

        if isinstance(viewer, int):
            to_add = self.viewers[viewer:viewer + 1]
        elif isinstance(viewer, slice):
            to_add = self.viewers[viewer]
        else:
            raise TypeError('Unable to find viewer at {}'.format(type(viewer)))

        for view in to_add:
            for v in visuals:
                view.add(v)

            if center:
                self.center_camera(view)

    def show(self):
        """ Show viewer. """
        self.canvas.show()

    def close(self):
        """ Close viewer. """
        if self == globals().get('viewer', None):
            globals().pop('viewer')
        self.canvas.close()

    def get_visuals(self, viewer):
        """ Returns list of all 3D visuals on given viewer. """
        if isinstance(viewer, int):
            viewer = self.viewers[viewer]

        return [v for v in viewer.children[0].children if isinstance(v, vp.scene.visuals.VisualNode)]

    def center_camera(self, viewer):
        """ Center camera on visuals. """
        if isinstance(viewer, int):
            viewer = self.viewers[viewer]

        viewer.camera.set_range()



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Module contains functions to plot neurons in 2D and 3D.
"""

import warnings

import plotly.offline
import numpy as np

from typing import Union, List, Optional

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import vispy

from .. import utils, config, core
from .vispy.viewer import Viewer
from .plotly.graph_objs import *

try:
    # Try setting vispy backend to PyQt5
    vispy.use(app='PyQt5')
except BaseException:
    pass

__all__ = ['plot3d']

logger = config.logger

if utils.is_jupyter():
    plotly.offline.init_notebook_mode(connected=True)


def plot3d(x: Union[core.NeuronObject,
                    core.Volume,
                    np.ndarray,
                    List[Union[core.NeuronObject, np.ndarray, core.Volume]]
                    ],
           **kwargs) -> Optional[Union[Viewer, dict]]:
    """ Generate 3D plot.

    Uses either `vispy <http://vispy.org>`_ or `plotly <http://plot.ly>`_.
    By default choice depends on context::

      terminal: vispy
      Jupyter: plotly

    See ``backend`` parameter on how to change default behavior.

    Parameters
    ----------

    x :               TreeNeuron/List| navis.Dotprops | navis.Volume | numpy.array
                        - ``numpy.array (N,3)`` is plotted as scatter plot
                        - multiple objects can be passed as list (see examples)
    backend :         'auto' | 'vispy' | 'plotly', default='auto'
                        - ``auto`` selects backend based on context: ``vispy``
                          for terminal and ``plotly`` for Jupyter environments.
                        - ``vispy`` uses OpenGL to generate high-performance
                          3D plots. Works in terminals.
                        - ``plotly`` generates 3D plots using WebGL. Works in
                          Jupyter notebooks. For Jupyter lab, you need to
                          have the Plotly lab extension installed.
    connectors :      bool, default=False
                      Plot connectors (e.g. synapses) if available.
    by_strahler :     bool, default=False
                      Will shade neuron(s) by strahler index.
    by_confidence :   bool, default=False
                      Will shade neuron(s) by arbor confidence.
    cn_mesh_colors :  bool, default=False
                      Plot connectors using mesh colors.
    clear3d :         bool, default=False
                      If True, canvas is cleared before plotting (only for
                      vispy).
    color :           None | str | tuple | list | dict, default=None (random)
                      Use single str (e.g. ``'red'``) or ``(r, g, b)`` tuple
                      to give all neurons the same color. Use ``list`` of
                      colors to assign colors: ``['red', (1, 0, 1), ...].
                      Use ``dict`` to map colors to neurons:
                      ``{uuid: (r, g, b), ...}``. RGB must be 0-255.
    use_neuron_color : bool, default=False
                      If True, will try using the ``.color`` attribute of
                      each neuron.
    width/height :    int, default=600
                      Use to define figure/window size.
    scatter_kws :     dict, optional
                      Use to modify scatter plots. Accepted parameters are
                        - ``size`` to adjust size of dots
                        - ``color`` to adjust color

    Plotly only

    title :           str, default=None
                      For plotly only! Change plot title.
    fig_autosize :    bool, default=False
                      For plotly only! Autoscale figure size.
                      Attention: autoscale overrides width and height
    plotly_inline :   bool, default=True
                      If True and you are in an Jupyter environment, will
                      render plotly plots inline. Else, will generate a
                      plotly figure dictionary that can be used to generate
                      a html with an embedded 3D plot.

    Returns
    --------
    If ``backend='vispy'``

        Opens a 3D window and returns :class:`navis.Viewer`.

    If ``backend='plotly'``

        Returns either ``None`` if you are in a Jupyter notebook (see
        ``plotly_inline`` parameter) or a ``fig`` dictionary to generate
        plotly 3D figure (see examples).

    See Also
    --------
    :class:`navis.Viewer`
        Interactive vispy 3D viewer. Makes it easy to add/remove/select
        objects.


    Examples
    --------
    >>> import navis

    In a Jupyter notebook using plotly as backend.

    >>> import plotly.offline
    >>> plotly.offline.init_notebook_mode()
    >>> nl = navis.example_neurons()
    >>> # Backend is automatically chosen but we can set it explicitly
    >>> # Plot inline
    >>> nl.plot3d(backend='plotly')
    >>> # Plot as separate html in a new window
    >>> fig = nl.plot3d(backend='plotly', plotly_inline=False)
    >>> plotly.offline.plot(fig)

    In a terminal using vispy as backend.

    >>> # Plot list of neurons
    >>> nl = navis.example_neurons()
    >>> v = navis.plot3d(nl, backend='vispy')
    >>> # Clear canvas
    >>> navis.clear3d()

    Some more advanced examples (using vispy here but also works with plotly).

    >>> # plot3d() can deal with combinations of objects
    >>> nl = navis.example_neurons()
    >>> vol = navis.example_volume('LH')
    >>> vol.color = (255, 0, 0, .5)
    >>> # This plots a neuronlists, a single neuron and a volume
    >>> navis.plot3d([nl[0:2], nl[3], vol])
    >>> # Pass kwargs
    >>> navis.plot3d(nl1, clear3d=True, by_strahler)

    """

    """
    TO-DO:
    - allow generic "plot_{}" arguments:
        - e.g. plot_nodes (default=True) or plot_connectors
        - either autodetect how to plot stuff (x, y, z required)
        - accept "{}_layout" parameter:
            {'size': either int/float or str (str=column)
             'as': 'scatter' | 'lineplot',
             'color': either color | str (column name)
             'colormap': if color is column, map to

              }
    """

    # Backend
    backend = kwargs.pop('backend', 'auto')
    allowed_backends = ['auto', 'vispy', 'plotly']
    if backend.lower() == 'auto':
        if utils.is_jupyter():
            backend = 'plotly'
        else:
            backend = 'vispy'
    elif backend.lower() not in allowed_backends:
        raise ValueError(f'Unknown backend "{backend}". '
                         f'Permitted: {".".join(allowed_backends)}.')

    if backend == 'vispy':
        return plot3d_vispy(x, **kwargs)
    else:
        return plot3d_plotly(x, **kwargs)


def plot3d_vispy(x, **kwargs):
    """ Plot3d() helper function to generate vispy 3D plots. This is just to
    improve readability. It's only purpose is to find the existing viewer
    or generate a new one.
    """

    # Parse objects to plot
    skdata, dotprops, volumes, points, visual = utils.parse_objects(x)

    # Check for allowed static parameters
    ALLOWED = {'color', 'c', 'colors', 'by_strahler', 'by_confidence',
               'cn_mesh_colors', 'linewidth', 'scatter_kws', 'synapse_layout',
               'dps_scale_vec', 'title', 'width', 'height',
               'auto_limits', 'autolimits', 'viewer', 'radius',
               'clear', 'clear3d', 'connectors', 'connectors_only'}

    # Check if any of these parameters are dynamic (i.e. attached data tables)
    notallowed = set(kwargs.keys()) - ALLOWED

    if any(notallowed):
        raise ValueError(f'Arguments "{",".join(notallowed)}" are not allowed '
                         'for plot3d using vispy. Allowed keyword '
                         f'arguments: {",".join(ALLOWED)}')

    scatter_kws = kwargs.pop('scatter_kws', {})

    if 'viewer' not in kwargs:
        # If does not exists yet, initialise a canvas object and make global
        if not getattr(config, 'primary_viewer', None):
            viewer = config.primary_viewer = Viewer()
        else:
            viewer = getattr(config, 'primary_viewer', None)
    else:
        viewer = kwargs.pop('viewer', getattr(config, 'primary_viewer'))

    # Make sure viewer is visible
    viewer.show()

    # We need to pop clear/clear3d to prevent clearing again later
    if kwargs.pop('clear3d', False) or kwargs.pop('clear', False):
        viewer.clear()

    if skdata:
        viewer.add(skdata, **kwargs)
    if not dotprops.empty:
        viewer.add(dotprops, **kwargs)
    if volumes:
        viewer.add(volumes, **kwargs)
    if points:
        viewer.add(points, scatter_kws=scatter_kws)

    return viewer


def plot3d_plotly(x, **kwargs):
    """
    Plot3d() helper function to generate plotly 3D plots. This is just to
    improve readability and structure of the code.
    """

    # Check for allowed static parameters
    ALLOWED = {'color', 'c', 'colors', 'by_strahler', 'by_confidence',
               'cn_mesh_colors', 'linewidth', 'scatter_kws', 'synapse_layout',
               'dps_scale_vec', 'title', 'width', 'height', 'fig_autosize',
               'plotly_inline',
               'connectors', 'connectors_only'}

    # Check if any of these parameters are dynamic (i.e. attached data tables)
    notallowed = set(kwargs.keys()) - ALLOWED

    if any(notallowed):
        raise ValueError(f'Arguments "{",".join(notallowed)}" are not allowed '
                         'for plot3d using vispy. Allowed keyword '
                         f'arguments: {",".join(ALLOWED)}')

    # Parse objects to plot
    skdata, dotprops, volumes, points, visual = utils.parse_objects(x)

    trace_data = []

    scatter_kws = kwargs.pop('scatter_kws', {})

    if skdata:
        trace_data += neuron2plotly(skdata, **kwargs)
    if not dotprops.empty:
        trace_data += dotprops2plotly(dotprops, **kwargs)
    if volumes:
        trace_data += volume2plotly(volumes, **kwargs)
    if points:
        trace_data += scatter2plotly(points, scatter_kws=scatter_kws)

    layout = layout2plotly(**kwargs)

    fig = dict(data=trace_data, layout=layout)

    if kwargs.get('plotly_inline', True) and utils.is_jupyter():
        plotly.offline.iplot(fig)
        return
    else:
        logger.info('Use plotly.offline.plot(fig, filename="3d_plot.html")'
                    ' to plot. Optimized for Google Chrome.')
        return fig



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Module contains functions to manage colours.
"""

import colorsys
import numbers

import matplotlib.colors as mcl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from typing import Union, List, Tuple, Optional, Dict, Any, Sequence, overload
from typing_extensions import Literal

from .. import core, config, utils

__all__ = ['generate_colors', 'prepare_connector_cmap', 'prepare_colormap',
           'eval_color', 'hex_to_rgb', 'vary_colors']

logger = config.logger

# Some definitions for mypy
RGB_color = Tuple[float, float, float]
RGBA_color = Tuple[float, float, float, float]
Str_color = str
ColorList = Sequence[Union[RGB_color, RGBA_color, Str_color]]
AnyColor = Union[RGB_color, RGBA_color, Str_color, ColorList]

def generate_colors(N: int,
                    color_space: Union[Literal['RGB'],
                                       Literal['Grayscale']] = 'RGB',
                    color_range: Union[Literal[1],
                                       Literal[255]] = 1
                    ) -> List[Tuple[float, float, float]]:
    """Divide colorspace into N evenly distributed colors.

    Returns
    -------
    colormap :  list
                [(r, g, b), (r, g, b), ...]

    """
    if N == 1:
        return [eval_color(config.default_color, color_range)]
    elif N == 0:
        return []

    # Make count_color an even number
    if N % 2 != 0:
        color_count = N + 1
    else:
        color_count = N

    colormap = []
    interval = 2 / color_count
    runs = int(color_count / 2)

    # Create first half with low brightness; second half with high brightness
    # and slightly shifted hue
    if color_space == 'RGB':
        for i in range(runs):
            # High brightness
            h = interval * i
            s = 1
            v = 1
            hsv = colorsys.hsv_to_rgb(h, s, v)
            colormap.append(tuple(v * color_range for v in hsv))

            # Lower brightness, but shift hue by half an interval
            h = interval * (i + 0.5)
            s = 1
            v = 0.5
            hsv = colorsys.hsv_to_rgb(h, s, v)
            colormap.append(tuple(v * color_range for v in hsv))
    elif color_space == 'Grayscale':
        h = 0
        s = 0
        for i in range(color_count):
            v = 1 / color_count * i
            hsv = colorsys.hsv_to_rgb(h, s, v)
            colormap.append(tuple(v * color_range for v in hsv))

    logger.debug(f'{color_count} random colors created: {colormap}')

    # Make sure we return exactly N colors
    return colormap[:N]


def map_colors(colors: Optional[Union[str,
                                      Tuple[float, float, float],
                                      Dict[Any, str],
                                      Dict[Any, Tuple[float, float, float]],
                                      List[Union[str,
                                                 Tuple[float, float, float]]
                                           ]
                                      ]
                                ],
               objects: Sequence[Any],
               color_range: Union[Literal[1], Literal[255]] = 255
               ) -> List[Tuple[float, float, float]]:
    """ Maps color(s) onto list of objects.

    Parameters
    ----------
    colors :        None | str | tuple | list-like | dict | None
                    Color(s) to map onto ``objects``. Can be::

                      str: e.g. "blue", "k" or "y"
                      tuple: (0, 0, 1), (0, 0, 0) or (0, 1, 1)
                      list-like of the above: [(0, 0, 1), 'r', 'k', ...]
                      dict mapping objects to colors: {object1: 'r',
                                                       object2: (1, 1, 1)}

                    If list-like or dict do not cover all ``objects``, will
                    fall back to ``navis.config.default_color``. If ``None``,
                    will generate evenly spread out colors.

    objects :       list-like
                    Object(s) to map color onto.
    color_range :   int, optional

    Returns
    -------
    list of tuples
                    Will match length of ``objects``.
    """

    if not utils.is_iterable(objects):
        objects = [objects]

    # If no colors, generate random colors
    if isinstance(colors, type(None)):
        if len(objects) == 1:
            return [eval_color(config.default_color, color_range)]
        return generate_colors(len(objects),
                               color_space='RGB',
                               color_range=color_range)

    # Bring colors in the right space
    colors = eval_color(colors, color_range=color_range)

    # Match them to objects
    if isinstance(colors, dict):
        # If dict, try mapping to objects
        if set(objects) - set(colors.keys()):
            logger.warning('Objects w/o colors - falling back to default.')
        return [colors.get(o, config.default_color) for o in objects]
    elif isinstance(colors, tuple):
        # If single color map to each object
        return [colors] * len(objects)
    elif isinstance(colors, list):
        # If list of correct length, map onto objets
        if len(colors) != len(objects):
            logger.warning('N colours does not match N objects.')
        miss = len(objects) - len(colors) if len(objects) > len(colors) else 0
        return colors[: len(objects)] + [config.default_color] * miss
    else:
        raise TypeError(f'Unable to interpret colors of type "{type(colors)}"')


def prepare_connector_cmap(x) -> Dict[str, Tuple[float, float, float]]:
    """Look for "label" or "type" column in connector tables and generates
    a color for every unique type. See ``navis.set_default_connector_colors``.

    Returns
    -------
    dict
            Maps type to color. Will be empty if no types.
    """

    if isinstance(x, (core.NeuronList, core.TreeNeuron)):
        connectors = x.get('connectors', None)

        if not isinstance(connectors, pd.DataFrame):
            unique: List[str] = []
        elif 'type' in connectors:
            unique = connectors.type.unique()
        elif 'label' in connectors:
            unique = connectors.label.unique()
        elif 'relation' in connectors:
            unique = connectors.relation.unique()
        else:
            unique = []
    else:
        unique = list(set(x))

    colors = config.default_connector_colors
    if isinstance(colors, (list, np.ndarray)):
        if len(unique) > len(colors):
            raise ValueError('Must define more default connector colors. See'
                             'navis.set_default_connector_colors')

        return {t: config.default_connector_colors[i] for i, t in enumerate(unique)}
    elif isinstance(colors, dict):
        miss = [l for l in unique if l not in colors]
        if miss:
            raise ValueError(f'Connector labels/types {",".join(miss)} are not'
                             ' defined in default connector colors. '
                             'See navis.set_default_connector_colors')
        return colors
    else:
        raise TypeError('config.default_color must be dict or iterable, '
                        f'not {type(config.default_color)}')


def prepare_colormap(colors, skdata=None, dotprops=None, volumes=None,
                     use_neuron_color=False, color_range=255):
    """ Maps color(s) to neuron/dotprop colorlists.
    """

    # Prepare dummies in case either no skdata, no dotprops or no volumes
    if isinstance(skdata, type(None)):
        skdata = core.NeuronList([])
    elif not isinstance(skdata, core.NeuronList):
        skdata = core.NeuronList((skdata))

    if isinstance(dotprops, type(None)):
        dotprops = core.Dotprops()
        dotprops['gene_name'] = []

    if isinstance(volumes, type(None)):
        volumes = np.array([])

    if not isinstance(volumes, np.ndarray):
        volumes = np.array(volumes)

    # Only dotprops and neurons REQUIRE a color
    # Volumes are second class citiziens here
    colors_required = skdata.shape[0] + dotprops.shape[0]

    if not colors_required and len(volumes) == 0:
        # If no neurons to plot, just return None
        # This happens when there is only a scatter plot
        return [None], [None], [None]

    # If no colors, generate random colors
    if isinstance(colors, type(None)):
        colors = []
        colors += generate_colors(colors_required,
                                  color_space='RGB',
                                  color_range=color_range)
        colors += [getattr(v, 'color', (1, 1, 1)) for v in volumes]

    # We need to parse once here to convert named colours to rgb
    colors = eval_color(colors, color_range=color_range)

    # In order to cater for duplicate skeleton IDs in skdata (e.g. from
    # splitting into fragments), we will not map skids to colors but instead
    # keep colors as a list. That way users can pass a simple list of colors.

    # If dictionary, map skids to dotprops gene names and neuron skeleton IDs
    dotprop_cmap = []
    neuron_cmap = []
    volumes_cmap = []
    dc = config.default_color
    if isinstance(colors, dict):
        # Try finding color first by neuron, then uuid and finally by name
        neuron_cmap = []
        for n in skdata:
            this_c = dc
            for k in [n, n.id, n.name]:
                if k in colors:
                    this_c = colors[k]
                    break
            neuron_cmap.append(this_c)

        dotprop_cmap = [colors.get(s, dc) for s in dotprops.gene_name.values]

        # Try finding color first by volume, then uuid and finally by name
        # If no color found, fall back to color property
        volumes_cmap = []
        for v in volumes:
            this_c = getattr(v, 'color', (.95, .95, .95, .1))
            for k in [v, v.id, getattr(v, 'name', None)]:
                if k and k in colors:
                    this_c = colors[k]
                    break
            volumes_cmap.append(this_c)
    elif isinstance(colors, mcl.Colormap):
        # Generate colors for neurons and dotprops
        neuron_cmap = [colors(i/len(skdata)) for i in range(len(skdata))]
        dotprop_cmap = [colors(i/dotprops.shape[0]) for i in range(dotprops.shape[0])]

        # Colormaps are not applied to volumes
        volumes_cmap = [getattr(v, 'color', (.95, .95, .95, .1)) for v in volumes]
    # If list of colors
    elif isinstance(colors, (list, tuple, np.ndarray)):
        # If color is a single color, convert to list
        if all([isinstance(elem, numbers.Number) for elem in colors]):
            # Generate at least one color
            colors = [colors] * max(colors_required, 1)
        elif len(colors) < colors_required:
            raise ValueError(f'Need colors for {colors_required} neurons/'
                             f'dotprops, got {len(colors)}')
        elif len(colors) > colors_required:
            logger.debug(f'More colors than required: got {len(colors)}, '
                         f'needed {colors_required}')

        if skdata.shape[0]:
            neuron_cmap = [colors.pop(0) for i in range(skdata.shape[0])]
        if dotprops.shape[0]:
            dotprop_cmap = [colors.pop(0) for i in range(dotprops.shape[0])]
        if volumes.shape[0]:
            # Volume have their own color property as fallback
            volumes_cmap = []
            for v in volumes:
                if colors:
                    volumes_cmap.append(colors.pop(0))
                else:
                    volumes_cmap.append(getattr(v, 'color', (.8, .8, .8, .2)))
    else:
        raise TypeError(f'Unable to parse colors of type "{type(colors)}"')

    # Override neuron cmap if we are supposed to use neuron colors
    if use_neuron_color:
        neuron_cmap = [getattr(n, 'color', config.default_color)
                       for i, n in enumerate(skdata)]

    # Make sure colour range checks out
    neuron_cmap = [eval_color(c, color_range=color_range)
                   for c in neuron_cmap]
    dotprop_cmap = [eval_color(c, color_range=color_range)
                    for c in dotprop_cmap]
    volumes_cmap = [eval_color(c, color_range=color_range)
                    for c in volumes_cmap]

    logger.debug('Neuron colormap: ' + str(neuron_cmap))
    logger.debug('Dotprops colormap: ' + str(dotprop_cmap))
    logger.debug('Volumes colormap: ' + str(volumes_cmap))

    return neuron_cmap, dotprop_cmap, volumes_cmap


def eval_color(x, color_range=255, force_alpha=False):
    """Helper to evaluate colors. Always returns tuples."""

    if color_range not in [1, 255]:
        raise ValueError('"color_range" must be 1 or 255')

    if isinstance(x, str):
        # Check if named color
        if mcl.is_color_like(x):
            c = mcl.to_rgb(x)
        # Assume it's a matplotlib color map
        else:
            try:
                c = plt.get_cmap(x)
            except ValueError:
                raise ValueError('Unable to interpret color "{}"'.format(x))
            except BaseException:
                raise
    elif isinstance(x, dict):
        return {k: eval_color(v, color_range=color_range) for k, v in x.items()}
    elif isinstance(x, (list, tuple, np.ndarray)):
        # If is this is not a list of RGB values:
        if any([not isinstance(elem, numbers.Number) for elem in x]):
            return [eval_color(c, color_range=color_range) for c in x]
        # If this is a single RGB color:
        c = x
    elif isinstance(x, type(None)):
        return None
    else:
        raise TypeError(f'Unable to interpret color of type "{type(x)}"')

    if not isinstance(c, mcl.Colormap):
        # Check if we need to convert
        if not any([v > 1 for v in c[:3]]) and color_range == 255:
            c = np.array(c, dtype=float)
            c[:3] = (c[:3] * 255).astype(int)
        elif any([v > 1 for v in c[:3]]) and color_range == 1:
            c = np.array(c, dtype=float)
            c[:3] = c[:3] / 255

        c = tuple(c)

    if force_alpha and len(c) == 3:
        c = (c[0], c[1], c[2], 1)

    return c


def hex_to_rgb(value: str) -> Tuple[int, int, int]:
    """Convert hex to rgb."""
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))  # type: ignore


def vary_colors(color: AnyColor,
                by_max: float = .1) -> np.ndarray:
    """Add small variance to color"""
    if isinstance(color, str):
        color = mcl.to_rgb(color)

    if not isinstance(color, np.ndarray):
        color = np.array(color)

    if color.ndim == 1:
        color = color.reshape(1, color.shape[0])

    variance = (np.random.randint(0, 100, color.shape) / 100) * by_max
    variance = variance - by_max / 2

    # We need to make sure color is array of floats
    color = color.astype(float)
    color[:, :3] = color[:, :3] + variance[:, :3]

    return np.clip(color, 0, 1)



""" Test suite for pymaid

This test suite requires a bunch of variables that need to either defined
as environment variables or in a config_test.py file

    #Catmaid server url
    server_url = ''

    #Http user
    http_user = ''

    #Http password
    http_pw = ''

    #Auth token
    token = ''

    # Test skeleton IDs
    test_skids = []

    # Test annotations
    test_annotations = []

    # Test CATMAID volume
    test_volume = ''

If you use environmental variables, give lists as comma-separated string.


Examples
--------

From terminal:
>>> python test_pymaid.py

From shell:
>>> import unittest
>>> import test_pymaid
>>> suite = unittest.TestLoader().loadTestsFromModule(test_pymaid)
>>> unittest.TextTestRunner().run(suite)

"""

import warnings
import os
import matplotlib as mpl
#if os.environ.get('DISPLAY', '') == '':
#    warnings.warn('No display found. Using template backend (nothing '
#                  'will show).')
#    mpl.use('template')
mpl.use('template')
import matplotlib.pyplot as plt

import unittest
import doctest

import datetime

import pymaid
import pandas as pd
import numpy as np
import networkx as nx

import importlib

try:
    import igraph
except BaseException:
    igraph = None
    warnings.warn('iGraph library not found. Will test only with NetworkX.')

# Silence module loggers
pymaid.set_loggers('ERROR')

# Deactivate progress bars
pymaid.set_pbars(hide=True)

# Try getting credentials from environment variable
required_variables = ['server_url', 'http_user', 'http_pw', 'token',
                      'test_skids', 'test_annotations', 'test_volume']

if False not in [v in os.environ for v in required_variables]:
    class conf:
        pass
    config_test = conf()

    for v in ['server_url', 'http_user', 'http_pw', 'token', 'test_volume']:
        setattr(config_test, v, os.environ[v])

    for v in ['test_skids', 'test_annotations']:
        setattr(config_test, v, os.environ[v].split(','))
else:
    missing = [v for v in required_variables if v not in os.environ]
    print('Missing environment variables:', ', '.join(missing))
    print('Falling back to config_test.py')
    try:
        import config_test
    except ImportError:
        raise ImportError('Unable to import configuration file.')


class TestModules(unittest.TestCase):
    """Test individual module import. """

    def test_imports(self):
        mods = ['morpho', 'core', 'plotting', 'graph', 'graph_utils', 'core',
                'connectivity', 'user_stats', 'cluster', 'resample',
                'intersect', 'fetch', 'scene3d']

        for m in mods:
            _ = importlib.import_module('pymaid.{}'.format(m))


class TestFetch(unittest.TestCase):
    """Test pymaid.fetch """

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            # if igraph:
            #    pymaid.config.use_igraph = True
            #    res2 = func(self, *args, **kwargs)
            #    self.assertEqual(res1, res2)
            return res1
        return wrapper

    @try_conditions
    def test_get_annotation_list(self):
        remote_annotation_list_url = self.rm._get_annotation_list()
        self.assertIn('annotations', self.rm.fetch(remote_annotation_list_url))

    @try_conditions
    def test_eval_skids(self):
        """ Test skeleton ID evaluation. """
        self.assertEqual(pymaid.eval_skids(
            ['12345', '67890'], remote_instance=self.rm), ['12345', '67890'])
        self.assertEqual(pymaid.eval_skids(
            [12345, 67890], remote_instance=self.rm), ['12345', '67890'])
        self.assertIsNotNone(pymaid.eval_skids(
            'annotation:{}'.format(config_test.test_annotations[0]),
            remote_instance=self.rm))

    @try_conditions
    def test_neuron_exists(self):
        self.assertIsInstance(pymaid.neuron_exists(
            config_test.test_skids[0], remote_instance=self.rm), bool)
        self.assertIsInstance(pymaid.neuron_exists(
            config_test.test_skids, remote_instance=self.rm), dict)

    @try_conditions
    def test_get_neuron_list(self):
        self.assertIsInstance(pymaid.get_neuron_list(
            node_count=20000, remote_instance=self.rm), list)

    @try_conditions
    def test_get_user_list(self):
        self.assertIsInstance(pymaid.get_user_list(
            remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_history(self):
        self.assertIsInstance(pymaid.get_history(
            remote_instance=self.rm), pd.Series)

    @try_conditions
    def test_get_annotated_skids(self):
        self.assertIsInstance(pymaid.get_skids_by_annotation(
            config_test.test_annotations[0], remote_instance=self.rm), list)

    @try_conditions
    def test_get_annotated(self):
        self.assertIsInstance(pymaid.get_annotated(
            config_test.test_annotations[0], remote_instance=self.rm),
                                                   pd.DataFrame)

    @try_conditions
    def test_get_annotations(self):
        self.assertIsInstance(pymaid.get_annotations(
            config_test.test_skids, remote_instance=self.rm), dict)
        self.assertIsInstance(pymaid.get_annotation_details(
            config_test.test_skids, remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_neuron(self):
        self.assertIsInstance(pymaid.get_neuron(config_test.test_skids,
                                                remote_instance=self.rm),
                              pymaid.CatmaidNeuronList)

    @try_conditions
    def test_get_neuron2(self):
        self.assertIsInstance(pymaid.get_arbor(
            config_test.test_skids[0], remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_treenode_table(self):
        self.assertIsInstance(pymaid.get_treenode_table(
            config_test.test_skids, remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_review(self):
        self.assertIsInstance(pymaid.get_review(
            config_test.test_skids, remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_volume(self):
        self.assertIsInstance(pymaid.get_volume(
            config_test.test_volume, remote_instance=self.rm), pymaid.Volume)

    @try_conditions
    def test_get_logs(self):
        self.assertIsInstance(pymaid.get_logs(
            remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_contributor_stats(self):
        self.assertIsInstance(pymaid.get_contributor_statistics(
            config_test.test_skids, remote_instance=self.rm), pd.Series)

    @try_conditions
    def test_get_partners(self):
        self.assertIsInstance(pymaid.get_partners(
            config_test.test_skids[0], remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_names(self):
        names = pymaid.get_names(
            config_test.test_skids, remote_instance=self.rm)
        self.assertIsInstance(names, dict)
        self.assertIsInstance(pymaid.get_skids_by_name(
            list(names.values()), remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_cn_table(self):
        self.assertIsInstance(pymaid.get_partners('annotation:%s' % config_test.test_annotations[
                              0], remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_connectors(self):
        cn = pymaid.get_connectors(
            config_test.test_skids, remote_instance=self.rm)
        self.assertIsInstance(cn, pd.DataFrame)
        self.assertIsInstance(pymaid.get_connector_details(
            cn.connector_id.tolist(), remote_instance=self.rm), pd.DataFrame)

    @try_conditions
    def test_get_connector_links(self):
        cn = pymaid.get_connector_links(config_test.test_skids,
                                        remote_instance=self.rm)
        self.assertIsInstance(cn, pd.DataFrame)

    @try_conditions
    def test_get_partners_in_volume(self):
        self.assertIsInstance(pymaid.get_partners_in_volume(config_test.test_skids[0],
                                                            config_test.test_volume),
                              pd.DataFrame)

    @try_conditions
    def test_node_details(self):
        n = pymaid.get_neuron(config_test.test_skids[0])
        self.assertIsInstance(pymaid.get_node_details(n.nodes.sample(100).node_id.values),
                              pd.DataFrame)

    @try_conditions
    def test_skid_from_treenode(self):
        n = pymaid.get_neuron(config_test.test_skids[0])
        self.assertIsInstance(pymaid.get_skid_from_treenode(n.nodes.iloc[0].node_id),
                              dict)

    @try_conditions
    def test_get_edges(self):
        self.assertIsInstance(pymaid.get_edges(config_test.test_skids),
                              pd.DataFrame)

    @try_conditions
    def test_connectors_between(self):
        self.assertIsInstance(pymaid.get_connectors_between(config_test.test_skids,
                                                            config_test.test_skids),
                              pd.DataFrame)

    @try_conditions
    def test_user_annotations(self):
        ul = pymaid.get_user_list()
        self.assertIsInstance(pymaid.get_user_annotations(ul.sample(1).iloc[0].id),
                              pd.DataFrame)

    @try_conditions
    def test_has_soma(self):
        self.assertIsInstance(pymaid.has_soma(config_test.test_skids[0]),
                              dict)

    @try_conditions
    def test_treenode_info(self):
        n = pymaid.get_neuron(config_test.test_skids[0])
        self.assertIsInstance(pymaid.get_treenode_info(n.nodes.node_id.values[0:50]),
                              pd.DataFrame)

    @try_conditions
    def test_treenode_tags(self):
        n = pymaid.get_neuron(config_test.test_skids[0])
        self.assertIsInstance(pymaid.get_node_tags(n.nodes.node_id.values[0:50],
                                                   node_type='TREENODE'),
                              dict)

    @try_conditions
    def test_review_details(self):
        self.assertIsInstance(pymaid.get_review_details(config_test.test_skids[0]),
                              pd.DataFrame)

    @try_conditions
    def test_find_neurons(self):
        self.assertIsInstance(pymaid.find_neurons(annotations=config_test.test_annotations),
                              pymaid.CatmaidNeuronList)

    @try_conditions
    def test_get_paths(self):
        paths, g = pymaid.get_paths(
            config_test.test_skids[0], config_test.test_skids[1], return_graph=True)
        self.assertIsInstance(g,
                              nx.Graph)
        self.assertIsInstance(paths,
                              list)

    @try_conditions
    def test_annotation_list(self):
        self.assertIsInstance(pymaid.get_annotation_list(),
                              pd.DataFrame)

    @try_conditions
    def test_url_to_coords(self):
        self.assertIsInstance(pymaid.url_to_coordinates((0, 0, 0), 1),
                              str)

    @try_conditions
    def test_get_segments(self):
        self.assertIsInstance(pymaid.get_segments(config_test.test_skids[0]),
                              list)

    @try_conditions
    def test_neurons_in_volume(self):
        self.assertIsInstance(pymaid.get_neurons_in_volume(config_test.test_volume),
                              list)

    @try_conditions
    def test_label_list(self):
        self.assertIsInstance(pymaid.get_label_list(),
                              pd.DataFrame)


class TestCore(unittest.TestCase):
    """Test pymaid.core """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.nl = pymaid.get_neuron('annotation:%s' % config_test.test_annotations[
            0], remote_instance=self.rm)

    @try_conditions
    def test_init(self):
        self.assertIsInstance(pymaid.CatmaidNeuron(
            config_test.test_skids[0]), pymaid.CatmaidNeuron)
        self.assertIsInstance(pymaid.CatmaidNeuronList(
            config_test.test_skids), pymaid.CatmaidNeuronList)

    @try_conditions
    def test_loaddata(self):
        # Test explicit loading
        nl = pymaid.CatmaidNeuron(config_test.test_skids[0])
        nl.get_skeleton()
        self.assertIsInstance(nl.nodes, pd.DataFrame)
        # Test implicit loading
        nl = pymaid.CatmaidNeuron(config_test.test_skids[0])
        self.assertIsInstance(nl.nodes, pd.DataFrame)

    @try_conditions
    def test_copy(self):
        self.assertIsInstance(self.nl[0].copy(), pymaid.CatmaidNeuron)
        self.assertIsInstance(self.nl.copy(), pymaid.CatmaidNeuronList)

    @try_conditions
    def test_summary(self):
        self.assertIsInstance(self.nl[0].summary(), pd.Series)
        self.assertIsInstance(self.nl.summary(), pd.DataFrame)

        self.assertIsInstance(self.nl.sum(), pd.Series)
        self.assertIsInstance(self.nl.mean(), pd.Series)

    @try_conditions
    def test_reload(self):
        self.nl.reload()
        self.assertIsInstance(self.nl, pymaid.CatmaidNeuronList)

    @try_conditions
    def test_graph_related(self):
        self.assertIsInstance(self.nl[0].graph, nx.Graph)
        self.assertIsInstance(self.nl[0].segments, list)

    @try_conditions
    def test_indexing(self):
        skids = self.nl.skeleton_id

        self.assertIsInstance(self.nl[0], pymaid.CatmaidNeuron)
        self.assertIsInstance(self.nl[:3], pymaid.CatmaidNeuronList)
        self.assertIsInstance(
            self.nl[self.nl.n_nodes > 1000], pymaid.CatmaidNeuronList)
        self.assertIsInstance(
            self.nl[list(skids[:-1])], pymaid.CatmaidNeuronList)
        self.assertIsInstance(self.nl - self.nl[0], pymaid.CatmaidNeuronList)
        self.assertIsInstance(
            self.nl[0] + self.nl[1], pymaid.CatmaidNeuronList)
        self.assertIsInstance(self.nl + self.nl[1], pymaid.CatmaidNeuronList)

        self.assertIsInstance(self.nl.sample(2), pymaid.CatmaidNeuronList)

    @try_conditions
    def test_neuron_attributes(self):
        attr = ['graph', 'simple', 'dps', 'annotations', 'partners',
                'review_status', 'nodes', 'connectors', 'presynapses',
                'postsynapses', 'gap_junctions', 'segments', 'soma',
                'root', 'tags', 'n_open_ends', 'n_end_nodes', 'n_connectors',
                'n_presynapses', 'n_postsynapses', 'cable_length']

        for a in attr:
            _ = getattr(self.nl[0], a)

    @try_conditions
    def test_neuron_functions(self):
        n = self.nl[0]
        slab = n.nodes[n.nodes.type == 'slab'].sample(1).iloc[0].node_id

        self.assertIsInstance(n.prune_distal_to(slab, inplace=False),
                              pymaid.CatmaidNeuron)
        self.assertIsInstance(n.prune_proximal_to(slab, inplace=False),
                              pymaid.CatmaidNeuron)
        self.assertIsInstance(n.prune_by_longest_neurite(inplace=False),
                              pymaid.CatmaidNeuron)

        self.assertIsInstance(n.reload(),
                              type(None))

    @try_conditions
    def test_swc_io(self):
        n = self.nl[0]

        n.to_swc('neuron.swc')

        self.assertIsInstance(pymaid.CatmaidNeuron.from_swc('neuron.swc'),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_json_io(self):
        n_string = pymaid.neuron2json(self.nl[:2])

        self.assertIsInstance(n_string, str)

        n = pymaid.json2neuron(n_string)

        self.assertIsInstance(n, pymaid.CatmaidNeuronList)

    @try_conditions
    def test_selection_io(self):
        self.nl.to_selection('selection.json')

        n = pymaid.CatmaidNeuronList.from_selection('selection.json')

        self.assertIsInstance(n, pymaid.CatmaidNeuronList)

    @try_conditions
    def test_has_annotation(self):
        self.assertIsInstance(self.nl.has_annotation(self.nl.annotations[0][0]),
                              pymaid.CatmaidNeuronList)

    @try_conditions
    def test_remove_duplicates(self):
        t = pymaid.CatmaidNeuronList([self.nl[0], self.nl[0], self.nl[0]])
        t2 = t.remove_duplicates(inplace=False)
        self.assertLess(len(t2), len(t))


class TestMorpho(unittest.TestCase):
    """ Test morphological operations """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.nl = pymaid.get_neuron(config_test.test_skids,
                                    remote_instance=self.rm)

    @try_conditions
    def test_downsampling(self):
        nl2 = self.nl.copy()
        nl2.downsample(4)
        self.assertLess(nl2.n_nodes.sum(), self.nl.n_nodes.sum())

    @try_conditions
    def test_resampling(self):
        nl2 = self.nl.resample(10000, inplace=False)
        self.assertNotEqual(nl2.n_nodes.sum(), self.nl.n_nodes.sum())

    @try_conditions
    def test_prune_by_strahler(self):
        nl2 = self.nl.prune_by_strahler(inplace=False, to_prune=1)
        self.assertLess(nl2.n_nodes.sum(), self.nl.n_nodes.sum())

    @try_conditions
    def test_axon_dendrite_split(self):
        self.assertIsInstance(pymaid.split_axon_dendrite(self.nl[0]),
                              pymaid.CatmaidNeuronList)

    @try_conditions
    def test_segregation_index(self):
        self.assertIsInstance(pymaid.segregation_index(self.nl[0]),
                              float)

    @try_conditions
    def test_bending_flow(self):
        self.assertIsInstance(pymaid.bending_flow(self.nl[0]),
                              type(None))

    @try_conditions
    def test_flow_centrality(self):
        self.assertIsInstance(pymaid.flow_centrality(self.nl[0]),
                              type(None))

    @try_conditions
    def test_stitching(self):
        self.assertIsInstance(pymaid.stitch_neurons(self.nl[:2],
                                                    method='NONE'),
                              pymaid.CatmaidNeuron)
        self.assertIsInstance(pymaid.stitch_neurons(self.nl[:2],
                                                    method='LEAFS'),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_averaging(self):
        self.assertIsInstance(pymaid.average_neurons(self.nl[:2]),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_tortuosity(self):
        self.assertIsInstance(pymaid.tortuosity(self.nl[0]),
                              float)
        self.assertIsInstance(pymaid.tortuosity(self.nl),
                              pd.DataFrame)

    @try_conditions
    def test_arbor_confidence(self):
        self.assertIsInstance(pymaid.arbor_confidence(self.nl[0],
                                                      inplace=False),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_calc_cable(self):
        self.assertIsNotNone(pymaid.calc_cable(self.nl[0]))

    @try_conditions
    def test_remove_branches(self):
        self.assertIsInstance(pymaid.remove_tagged_branches(self.nl[0],
                                                            'not a branch',
                                                            how='segment',
                                                            preserve_connectors=True,
                                                            inplace=False),
                              pymaid.CatmaidNeuron)

        self.assertIsInstance(pymaid.remove_tagged_branches(self.nl[0],
                                                            'not a branch',
                                                            how='distal',
                                                            preserve_connectors=True,
                                                            inplace=False),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_despike_neuron(self):
        self.assertIsInstance(pymaid.despike_neuron(self.nl[0],
                                                    inplace=False),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_guess_radius(self):
        self.assertIsInstance(pymaid.guess_radius(self.nl[0],
                                                  inplace=False),
                              pymaid.CatmaidNeuron)


class TestGraphs(unittest.TestCase):
    """Test pymaid.graph and pymaid.graph_utils """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.nl = pymaid.get_neuron(config_test.test_skids[0:2],
                                    remote_instance=self.rm)

        self.n = self.nl[0]
        self.n.reroot(self.n.soma)

        # Get some random leaf node
        self.leaf_id = self.n.nodes[self.n.nodes.type == 'end'].sample(
            1).iloc[0].node_id
        self.slab_id = self.n.nodes[self.n.nodes.type == 'slab'].sample(
            1).iloc[0].node_id

    @try_conditions
    def test_reroot(self):
        self.assertIsNotNone(self.n.reroot(self.leaf_id, inplace=False))
        self.assertIsNotNone(self.nl.reroot(self.nl.soma, inplace=False))

    @try_conditions
    def test_distal_to(self):
        self.assertTrue(pymaid.distal_to(self.n, self.leaf_id, self.n.root))
        self.assertFalse(pymaid.distal_to(self.n, self.n.root, self.leaf_id))

    @try_conditions
    def test_distance(self):
        leaf_id = self.n.nodes[self.n.nodes.type == 'end'].iloc[0].node_id

        self.assertIsNotNone(pymaid.dist_between(self.n,
                                                 leaf_id,
                                                 self.n.root))
        self.assertIsNotNone(pymaid.dist_between(self.n,
                                                 self.n.root,
                                                 leaf_id))

    @try_conditions
    def test_find_bp(self):
        self.assertIsNotNone(pymaid.find_main_branchpoint(self.n,
                                                          reroot_to_soma=False))

    @try_conditions
    def test_split_fragments(self):
        self.assertIsNotNone(pymaid.split_into_fragments(self.n,
                                                         n=2,
                                                         reroot_to_soma=False))

    @try_conditions
    def test_longest_neurite(self):
        self.assertIsNotNone(pymaid.longest_neurite(self.n,
                                                    n=2,
                                                    reroot_to_soma=False))

    @try_conditions
    def test_cut_neuron(self):
        dist, prox = pymaid.cut_neuron(
            self.n,
            self.slab_id,
        )
        self.assertNotEqual(dist.nodes.shape, prox.nodes.shape)

        # Make sure dist and prox check out
        self.assertTrue(pymaid.distal_to(self.n, dist.root, prox.root))

    @try_conditions
    def test_subset(self):
        self.assertIsInstance(pymaid.subset_neuron(self.n,
                                                   self.n.segments[0]),
                              pymaid.CatmaidNeuron)

    @try_conditions
    def test_node_sorting(self):
        self.assertIsInstance(pymaid.node_label_sorting(self.n),
                              list)


class TestConnectivity(unittest.TestCase):
    """Test pymaid.plotting """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.n = pymaid.get_neuron(config_test.test_skids[0],
                                   remote_instance=self.rm)

        self.cn_table = pymaid.get_partners(config_test.test_skids[0],
                                            remote_instance=self.rm)

        self.nB = pymaid.get_neuron(self.cn_table.iloc[0].skeleton_id,
                                    remote_instance=self.rm)

        self.adj = pymaid.adjacency_matrix(
            self.cn_table[self.cn_table.relation == 'upstream'].iloc[:10].skeleton_id.values)

    @try_conditions
    def test_adjacency_matrix(self):
        self.assertIsInstance(self.adj, pd.DataFrame)

    @try_conditions
    def test_adjacency_matrix2(self):
        nl = pymaid.get_neurons(
            self.cn_table[self.cn_table.relation == 'upstream'].iloc[:10].skeleton_id.values)
        self.assertIsInstance(pymaid.adjacency_matrix(nl, use_connectors=True),
                              pd.DataFrame)

    @try_conditions
    def test_group_matrix(self):
        gr_adj = pymaid.group_matrix(self.adj,
                                     row_groups={n: 'group1' for n in self.adj.index.values})
        self.assertIsInstance(gr_adj, pd.DataFrame)

    @try_conditions
    def test_connectivity_filter(self):
        dist, prox = pymaid.cut_neuron(
            self.n, pymaid.find_main_branchpoint(self.n))

        vol = pymaid.get_volume(config_test.test_volume)

        # Connectivity table by neuron
        self.assertIsInstance(pymaid.filter_connectivity(self.cn_table, prox),
                              pd.DataFrame)
        # Adjacency matrix by neuron
        self.assertIsInstance(pymaid.filter_connectivity(self.adj, prox),
                              pd.DataFrame)

        # Connectivity table by volume
        self.assertIsInstance(pymaid.filter_connectivity(self.cn_table, vol),
                              pd.DataFrame)
        # Adjacency matrix by volume
        self.assertIsInstance(pymaid.filter_connectivity(self.adj, vol),
                              pd.DataFrame)

    @try_conditions
    def test_calc_overlap(self):
        self.assertIsInstance(pymaid.cable_overlap(self.n, self.nB),
                              pd.DataFrame)

    @try_conditions
    def test_pred_connectivity(self):
        self.assertIsInstance(pymaid.predict_connectivity(self.n,
                                                          self.nB,
                                                          remote_instance=self.rm),
                              pd.DataFrame)

    @try_conditions
    def test_cn_table_from_connectors(self):
        self.assertIsInstance(pymaid.cn_table_from_connectors(self.n,
                                                              remote_instance=self.rm),
                              pd.DataFrame)

    @try_conditions
    def test_adjacency_from_connectors(self):
        nl = pymaid.get_neurons(
            self.cn_table[self.cn_table.relation == 'upstream'].iloc[:10].skeleton_id.values)
        self.assertIsInstance(pymaid.adjacency_from_connectors(nl,
                                                               remote_instance=self.rm),
                              pd.DataFrame)


class TestCluster(unittest.TestCase):
    """Test pymaid.cluster """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

    @try_conditions
    def test_connectivity_cluster(self):
        self.assertIsInstance(pymaid.cluster_by_connectivity(config_test.test_skids),
                              pymaid.ClustResults)

    @try_conditions
    def test_synapse_cluster(self):
        self.assertIsInstance(pymaid.cluster_by_synapse_placement(config_test.test_skids),
                              pymaid.ClustResults)

    @try_conditions
    def test_clustresults(self):
        res = pymaid.cluster_by_connectivity(config_test.test_skids)

        self.assertIsNotNone(res.linkage, np.ndarray)
        self.assertIsNotNone(res.leafs)
        self.assertIsNotNone(res.cophenet)
        self.assertIsNotNone(res.agg_coeff)
        res.to_selection()
        self.assertIsInstance(res.get_colormap(k=2), dict)
        self.assertIsInstance(res.get_clusters(k=2), list)


class TestPlot(unittest.TestCase):
    """Test pymaid.plotting """

    def try_conditions(func):
        """Runs each test under various conditions and asserts that results
        are always the same."""

        def wrapper(self, *args, **kwargs):
            pymaid.config.use_igraph = False
            res1 = func(self, *args, **kwargs)
            if igraph:
                pymaid.config.use_igraph = True
                res2 = func(self, *args, **kwargs)
                self.assertEqual(res1, res2)
            return res1
        return wrapper

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.nl = pymaid.get_neuron(config_test.test_skids,
                                    remote_instance=self.rm)

        self.vol = pymaid.get_volume(config_test.test_volume)

    """
    @try_conditions
    def test_plot3d_plotly(self):
        self.assertIsNotNone(self.nl.plot3d(backend='plotly'))
        self.assertIsNotNone(pymaid.plot3d(self.nl, backend='plotly'))
        self.assertIsNotNone(pymaid.plot3d(
            [self.nl, self.vol], backend='plotly'))

    @try_conditions
    def test_plot3d_vispy(self):
        # self.assertIsNotNone(self.nl.plot3d(backend='vispy'))
        # pymaid.close3d()
        # self.assertIsNotNone(pymaid.plot3d(self.nl, backend='vispy'))
        # pymaid.close3d()
        # self.assertIsNotNone(pymaid.plot3d([self.nl, self.vol],
        #                                   backend='vispy'))
        # pymaid.close3d()
        self.assertIsNotNone(pymaid.plotting._neuron2vispy(self.nl))
        self.assertIsNotNone(pymaid.plotting._volume2vispy(self.vol))
    """

    @try_conditions
    def test_plot2d(self):
        plt.close()
        self.assertIsNotNone(self.nl.plot2d(method='2d'))
        plt.close()
        self.assertIsNotNone(self.nl.plot2d(method='3d_complex'))
        plt.close()
        self.assertIsNotNone(self.nl.plot2d(method='3d'))
        plt.close()

    def tearDown(self):
        plt.close()


class TestTiles(unittest.TestCase):
    """Test pymaid.tiles """

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(config_test.server_url,
                                         config_test.http_user,
                                         config_test.http_pw,
                                         config_test.token)

    def test_tiles(self):
        from navis import tiles
        # Generate the job
        job = tiles.LoadTiles([119000, 119500, 36000, 36500, 4050],
                              stack_id=5,
                              coords='PIXEL')
        # Load, stich and crop the required EM image tiles
        job.load_in_memory()
        # Render image
        ax = job.render_im(slider=False, figsize=(12, 12))
        # Add treenodes
        job.render_nodes(ax, treenodes=True, connectors=False)
        # Add scalebar
        job.scalebar(size=1000, ax=ax, label=False)
        # Show
        plt.close()

    def tearDown(self):
        plt.close()


class TestUserStats(unittest.TestCase):
    """Test pymaid.user_stats """

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(
            config_test.server_url,
            config_test.http_user,
            config_test.http_pw,
            config_test.token)

        self.n = pymaid.get_neuron(config_test.test_skids[0],
                                   remote_instance=self.rm)

    def test_time_invested(self):
        ds = self.n.downsample(20, inplace=False)
        self.assertIsInstance(pymaid.get_time_invested(ds,
                                                       mode='SUM',
                                                       remote_instance=self.rm),
                              pd.DataFrame)
        self.assertIsInstance(pymaid.get_time_invested(ds,
                                                       mode='ACTIONS',
                                                       remote_instance=self.rm),
                              pd.DataFrame)
        self.assertIsInstance(pymaid.get_time_invested(ds,
                                                       mode='OVER_TIME',
                                                       remote_instance=self.rm),
                              pd.DataFrame)

    def test_team_contributions(self):
        ds = self.n.downsample(20, inplace=False)
        ul = pymaid.get_user_list().set_index('id')
        teams = {'test_team' : [ul.loc[u, 'login'] for u in ds.nodes.creator_id.unique()]}
        self.assertIsInstance(pymaid.get_team_contributions(teams,
                                                            neurons=ds,
                                                            remote_instance=self.rm),
                              pd.DataFrame)

    def test_user_contributions(self):
        self.assertIsInstance(pymaid.get_user_contributions(
            config_test.test_skids, remote_instance=self.rm), pd.DataFrame)

    def test_user_actions(self):
        # Get a day on which a random user has done some work
        last_week = datetime.date.today() - datetime.timedelta(days=7)
        h = pymaid.get_history(start_date=last_week,
                               end_date=datetime.date.today())
        u = h.treenodes.loc[h.treenodes.sum(axis=1) > 10].sample(1).index[0]

        user_actions = pymaid.get_user_actions(users=u, start_date=last_week,
                                               end_date=datetime.date.today())

        self.assertIsInstance(user_actions, pd.DataFrame)


class TestExamples(unittest.TestCase):
    """Test pymaid.tiles """

    def setUp(self):
        self.rm = pymaid.CatmaidInstance(config_test.server_url,
                                         config_test.http_user,
                                         config_test.http_pw,
                                         config_test.token)

    def test_fetch_examples(self):
        for func in pymaid.fetch.__all__:
            f = getattr(pymaid.fetch, func)
            doctest.run_docstring_examples(f, globals(), name=f)

    def test_user_stats_examples(self):
        for func in pymaid.user_stats.__all__:
            f = getattr(pymaid.user_stats, func)
            doctest.run_docstring_examples(f, globals(), name=f)


if __name__ == '__main__':
    unittest.main()



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import os
import colorsys
import math

import numpy as np
import pandas as pd
import scipy.cluster.hierarchy
import scipy.spatial

from typing import Union, Optional, List, Dict, Sequence
from typing_extensions import Literal

from concurrent.futures import ThreadPoolExecutor

from ..core.neurons import TreeNeuron
from ..core.neuronlist import NeuronList
from .. import plotting, utils, config

# Set up logging
logger = config.logger

NeuronObject = Union[TreeNeuron, NeuronList]

__all__ = sorted(['cluster_by_connectivity', 'cluster_by_synapse_placement',
                  'cluster_xyz', 'ClustResults'])


def cluster_by_connectivity(adjacency: Union[pd.DataFrame, np.ndarray],
                            similarity: Union[Literal['matching_index'],
                                              Literal['matching_index_synapses'],
                                              Literal['matching_index_weighted_synapses'],
                                              Literal['vertex'],
                                              Literal['vertex_normalized']
                                              ] = 'vertex_normalized',
                            threshold: int = 1,
                            cluster_kws: dict = {}) -> 'ClustResults':
    """ Calculate connectivity similarity.

    This functions offers a selection of metrics to compare connectivity:

    .. list-table::
       :widths: 15 75
       :header-rows: 1

       * - Metric
         - Explanation
       * - matching_index
         - Number of shared partners divided by total number of partners.
       * - matching_index_synapses
         - Number of shared synapses (i.e. number of connections from/onto the
           same partners) divided by total number of synapses. Attention: this
           metric is tricky when there is a disparity of total number of
           connections between neuron A and B. For example, consider 100/200
           and 1/50 shared/total synapse: 101/250 results in a fairly high
           matching index of 0.404.
       * - matching_index_weighted_synapses
         - Similar to *matching_index_synapses* but slightly less prone to
           above mentioned error as it uses the percentage of shared synapses:

           .. math::

               S = \\frac{\\text{NeuronA}_{\\text{ shared synapses}}}{\\text{NeuronA}_{\\text{ total synapses}}} \\times \\frac{\\text{NeuronB}_{\\text{ shared synapses}}}{\\text{NeuronB}_{\\text{ total synapses}}}

       * - vertex
         - Matching index that rewards shared and punishes non-shared partners.
           Based on
           `Jarrell et al., 2012 <http://science.sciencemag.org/content/337/6093/437>`_:

           .. math::

               f(x,y) = min(x,y) - C1 \\times max(x,y) \\times \\exp(-C2 * min(x,y))

           Final score is the sum of :math:`f(x,y)` over all edges x, y between
           neurons A+B and their partners. C1 determines how negatively a case
           where one edge is much stronger than another is punished. C2
           determines the point where the similarity switches from negative to
           positive. C1 and C2 default to 0.5 and 1, respectively, but can be
           changed by passing them in a dictionary as `cluster_kws`.
       * - vertex_normalized
         - This is *vertex* similarity normalized by the lowest (hypothetical
           total dissimilarity) and highest (all edge weights the same)
           achievable score.


    Parameters
    ----------
    adjacency :         pandas DataFrame | numpy array
                        Adjacency matrix. Does not have to be symmetrical.
                        Will calculate similarity for all row neurons using
                        the column neurons as observations.
    similarity :        'matching_index' | 'matching_index_synapses' | 'matching_index_weighted_synapses' | 'vertex' | 'vertex_normalized', optional
                        Metric used to compare connectivity. See notes for
                        detailed explanation.

    Returns
    -------
    :class:`navis.ClustResults`
                         Custom cluster results class holding the distance
                         matrix and contains wrappers e.g. to plot dendograms.
    """

    ALLOWED_METHODS = ['matching_index', 'matching_index_synapses',
                       'matching_index_weighted_synapses', 'vertex',
                       'vertex_normalized']
    if similarity.lower() not in ALLOWED_METHODS:
        raise ValueError(f'"method" must be either: {", ".join(ALLOWED_METHODS)}')

    if isinstance(adjacency, np.ndarray):
        adjacency = pd.DataFrame(adjacency)

    if not isinstance(adjacency, pd.DataFrame):
        raise TypeError(f'Expected DataFrame, got "{type(adjacency)}"')

    neurons = adjacency.index

    # Prepare empty matching scores
    matching_scores = pd.DataFrame(np.zeros((len(neurons), len(neurons))),
                                   index=neurons, columns=neurons)

    if similarity.lower() in ['vertex_normalized', 'vertex']:
        vertex_score = True
    else:
        vertex_score = False

    combinations = [(nA, nB, adjacency, threshold,
                     vertex_score, cluster_kws) for nA in neurons for nB in neurons]

    with ThreadPoolExecutor(max_workers=max(1, os.cpu_count())) as e:
        futures = e.map(_unpack_connectivity_helper, combinations)

        matching_indices = [n for n in config.tqdm(futures,
                                                   total=len(combinations),
                                                   desc='Calculating',
                                                   disable=config.pbar_hide,
                                                   leave=config.pbar_leave)]

    for i, v in enumerate(combinations):
            matching_scores.loc[v[0], v[1]] = matching_indices[i][similarity]

    results = ClustResults(matching_scores, mat_type='similarity')

    return results


def _unpack_connectivity_helper(x):
    """Helper function to unpack values from pool"""
    return _calc_connectivity_matching_index(neuronA=x[0],
                                             neuronB=x[1],
                                             connectivity=x[2],
                                             syn_threshold=x[3],
                                             vertex_score=x[4], **x[5])


def _calc_connectivity_matching_index(neuronA: int,
                                      neuronB: int,
                                      connectivity: pd.DataFrame,
                                      syn_threshold: int = 1,
                                      vertex_score: bool = True,
                                      **kwargs) -> Dict[str,
                                                        Union[float, int]]:
    """ Calculates and returns various matching indices between two neurons.

    Parameters
    ----------
    neuronA :         skeleton ID
    neuronB :         skeleton ID
    connectivity :    pandas DataFrame
                      Connectivity data as provided by :func:`navis.get_partners`.
    syn_threshold :   int, optional
                      Min number of synapses for a connection to be considered.
                      Default = 1
    vertex_score :    bool, optional
                      If False, no vertex score is returned (much faster!).
                      Default = True

    Returns
    -------
    dict
                      Containing all initially described matching indices

    Notes
    -----
    |matching_index =           Number of shared partners divided by total number
    |                           of partners

    |matching_index_synapses =  Number of shared synapses divided by total number
    |                           of synapses. Attention! matching_index_synapses
    |                           is tricky, because if neuronA has lots of
    |                           connections and neuronB only little, they will
    |                           still get a high matching index.
    |                           E.g. 100 of 200 / 1 of 50 = 101/250
    |                           -> ``matching index = 0.404``

    |matching_index_weighted_synapses = Similar to matching_index_synapses but
    |                           slightly less prone to above mentioned error:
    |                           % of shared synapses A * % of shared synapses
    |                           B * 2 / (% of shared synapses A + % of shared
    |                           synapses B)
    |                           -> value will be between 0 and 1; if one neuronB
    |                           has only few connections (percentage) to a shared
    |                           partner, the final value will also be small

    |vertex_normalized =        Matching index that rewards shared and punishes
    |                           non-shared partners. Vertex similarity based on
    |                           Jarrell et al., 2012:
    |                           f(x,y) = min(x,y) - C1 * max(x,y) * e^(-C2 * min(x,y))
    |                           x,y = edge weights to compare
    |                           vertex_similarity is the sum of f over all vertices
    |                           C1 determines how negatively a case where one edge
    |                           is much stronger than another is punished
    |                           C2 determines the point where the similarity
    |                           switches from negative to positive

    """
    # Subset to neurons that actually connect with our neuron pair
    this_cn = connectivity.loc[[neuronA, neuronB]]

    # Get all neurons that connect to either A or B
    total = this_cn.loc[:, (this_cn >= syn_threshold).sum(axis=0) >= 1]
    n_total = total.shape[1]

    # Get all neurons that connect to both A and B
    shared = this_cn.loc[:, (this_cn >= syn_threshold).sum(axis=0) == 2]
    n_shared = shared.shape[1]

    shared_sum = shared.sum(axis=1)
    n_synapses_sharedA = shared_sum[neuronA]
    n_synapses_sharedB = shared_sum[neuronB]

    total_sum = total.sum(axis=1)
    n_synapses_totalA = total_sum[neuronA]
    n_synapses_totalB = total_sum[neuronB]

    # If neuronA == neuronB, above counts will not be single values but
    # DataFrames -> we'll simply have to get the first values
    if neuronA == neuronB:
        n_synapses_sharedA = n_synapses_sharedA.values[0]
        n_synapses_sharedB = n_synapses_sharedB.values[0]

        n_synapses_totalA = n_synapses_totalA.values[0]
        n_synapses_totalB = n_synapses_totalB.values[0]

    # Sum up
    n_synapses_shared = n_synapses_sharedA + n_synapses_sharedB
    n_synapses_total = n_synapses_totalA + n_synapses_totalB

    # Vertex similarity based on Jarrell et al., 2012
    # f(x,y) = min(x,y) - C1 * max(x,y) * e^(-C2 * min(x,y))
    # x,y = edge weights to compare
    # vertex_similarity is the sum of f over all vertices
    # C1 determines how negatively a case where one edge is much stronger than
    #    another is punished
    # C2 determines the point where the similarity switches from negative to
    # positive
    C1 = kwargs.get('C1', 0.5)
    C2 = kwargs.get('C2', 1)
    similarity_indices: Dict[str, Union[float, int]] = {}

    if vertex_score:
        # Get min and max between both neurons
        this_max = np.max(this_cn, axis=0)
        this_min = np.min(this_cn, axis=0)

        # The max possible score is when both synapse counts are the same:
        # in which case score = max(x,y) - C1 * max(x,y) * e^(-C2 * max(x,y))
        max_score = this_max - C1 * this_max * np.exp(- C2 * this_max)

        # The smallest possible score is when either synapse count is 0:
        # in which case score = -C1 * max(a,b)
        min_score = -C1 * this_max

        # Implement: f(x,y) = min(x,y) - C1 * max(x,y) * e^(-C2 * min(x,y))
        v_sim = this_min - C1 * this_max * np.exp(- C2 * this_min)

        # Sum over all partners
        vertex_similarity = v_sim.sum()

        similarity_indices['vertex'] = vertex_similarity

        try:
            similarity_indices['vertex_normalized'] = (
                vertex_similarity - min_score.sum()) / (max_score.sum() - min_score.sum())
        except BaseException:
            similarity_indices['vertex_normalized'] = 0

    if n_total != 0:
        similarity_indices['matching_index'] = n_shared / n_total
        similarity_indices['matching_index_synapses'] = n_synapses_shared / n_synapses_total
        if n_synapses_sharedA != 0 and n_synapses_sharedB != 0:
            similarity_indices['matching_index_weighted_synapses'] = (
                n_synapses_sharedA / n_synapses_totalA) * (n_synapses_sharedB / n_synapses_totalB)
        else:
            # If no shared synapses at all:
            similarity_indices['matching_index_weighted_synapses'] = 0
    else:
        similarity_indices['matching_index'] = 0
        similarity_indices['matching_index_synapses'] = 0
        similarity_indices['matching_index_weighted_synapses'] = 0

    return similarity_indices


def _unpack_synapse_helper(x):
    """Helper function to unpack values from pool."""
    return _calc_synapse_similarity(x[0], x[1], x[2], x[3], x[4])


def _calc_synapse_similarity(cnA: pd.DataFrame,
                             cnB: pd.DataFrame,
                             sigma: int = 2,
                             omega: int = 2,
                             restrict_cn: Optional[List[str]] = None
                             ) -> float:
    """ Calculates synapses similarity score.

    Synapse similarity score is calculated by calculating for each synapse of
    neuron A: (1) the distance to the closest (eucledian) synapse in neuron B
    and (2) comparing the synapse density around synapse A and B. This is type
    sensitive: presynapses will only be matched with presynapses, post with
    post, etc. The formula is described in Schlegel et al., eLife (2017).

    Parameters
    ----------
    (cnA, cnB) :    Connector tables.
    sigma :         int, optional
                    Distance that is considered to be "close".
    omega :         int, optional
                    Radius over which to calculate synapse density.

    Returns
    -------
    synapse_similarity_score

    """

    all_values = []

    # Get the connector types that we want to compare between neuron A and B
    if isinstance(restrict_cn, type(None)):
        # If no restrictions, get all cn types in neuron A
        cn_to_check = cnA.type.unique()
    else:
        # Intersect restricted connectors and actually available types
        cn_to_check = set(cnA.type.unique()) & set(restrict_cn)

    # Iterate over all types of connectors
    for r in cn_to_check:
        # Skip if either neuronA or neuronB don't have this synapse type
        if cnB[cnB.type == r].empty:
            all_values += [0] * cnA[cnA.type == r].shape[0]
            continue

        # Get inter-neuron matrix
        dist_mat = scipy.spatial.distance.cdist(cnA[cnA.type == r][['x', 'y', 'z']],
                                                cnB[cnB.type == r][['x', 'y', 'z']])

        # Get index of closest synapse in neuron B
        closest_ix = np.argmin(dist_mat, axis=1)

        # Get closest distances
        closest_dist = dist_mat.min(axis=1)

        # Get intra-neuron matrices for synapse density checking
        distA = scipy.spatial.distance.pdist(
            cnA[cnA.type == r][['x', 'y', 'z']])
        distA = scipy.spatial.distance.squareform(distA)
        distB = scipy.spatial.distance.pdist(
            cnB[cnB.type == r][['x', 'y', 'z']])
        distB = scipy.spatial.distance.squareform(distB)

        # Calculate number of synapses closer than OMEGA. This does count itself!
        closeA = (distA <= omega).sum(axis=1)
        closeB = (distB <= omega).sum(axis=1)

        # Now calculate the scores over all synapses
        for a in range(distA.shape[0]):
            this_synapse_value = math.exp(-1 * math.fabs(closeA[a] - closeB[closest_ix[a]]) / (
                closeA[a] + closeB[closest_ix[a]])) * math.exp(-1 * (closest_dist[a]**2) / (2 * sigma**2))
            all_values.append(this_synapse_value)

    score = sum(all_values) / len(all_values)

    return score


def cluster_by_synapse_placement(x: Union[NeuronList,
                                          Dict[Union[str, int], pd.DataFrame]],
                                 sigma: int = 2,
                                 omega: int = 2,
                                 mu_score: bool = True,
                                 restrict_cn: Optional[List[str]] = None
                                 ) -> 'ClustResults':
    """ Clusters neurons based on their synapse placement.

    Distances score is calculated by calculating for each synapse of
    neuron A: (1) the distance to the closest (eucledian) synapse in neuron B
    and (2) comparing the synapse density around synapse A and B.
    This is type-sensitive: presynapses will only be matched with presynapses,
    post with post, etc. The formula is described in
    `Schlegel et al., eLife (2017) <https://elifesciences.org/articles/16799>`_:

    .. math::

        f(i_{s},j_{k}) = \\exp(\\frac{-d^{2}_{sk}}{2\\sigma^{2}}) \\exp(\\frac{|n(i_{s})-n(j_{k})|}{n(i_{s})+n(j_{k})})

    The synapse similarity score for neurons i and j being the average
    of :math:`f(i_{s},j_{k})` over all synapses s of i. Synapse k is the
    closest synapse of the same sign (pre/post) in neuron j to synapse s.
    :math:`d^{2}_{sk}` is the eucledian distance between these distances.
    Variable :math:`\\sigma` (``sigma``) determines what distance between
    s and k is considered "close". :math:`n(i_{s})` and :math:`n(j_{k})` are
    defined as the number of synapses of neuron i/j that are within given
    radius :math:`\\omega` (``omega``) of synapse s and j, respectively (same
    sign only). This esnures that in cases of a strong disparity between
    :math:`n(i_{s})` and :math:`n(j_{k})`, the synapse similarity will be
    close to zero even if the distance between s and k is very small.


    Parameters
    ----------
    x :                 NeuronList | Dict of pandas.DataFrame
                        Neurons to compare. Dict must map
                        ``{neuron_name: connector_table}``. If NeuronList,
                        each neuron must have ``.connectors`` connector tables.
                        Connector tables must have `x`, `y`, `y` columns. If
                        a `type`/`relation`/`label` columns is present, will
                        use this to compare connectors only within type.
    sigma :             int, optional
                        Distance between synapses that is considered to be
                        "close".
    omega :             int, optional
                        Radius over which to calculate synapse density.
    mu_score :          bool, optional
                        If True, score is calculated as mean between A->B and
                        B->A comparison.
    restrict_cn :       int | list | None, optional
                        Restrict to given connector types. Must map to
                        a `type`, `relation` or `label` column in the
                        connector tables.
                        If None, will use all connector types. Use either
                        single integer or list. E.g. ``restrict_cn=[0, 1]``
                        to use only pre- and postsynapses.

    Returns
    -------
    :class:`~navis.ClustResults`
                Object that contains distance matrix and methods to plot
                dendrograms.

    """

    if isinstance(x, dict):
        if any([not isinstance(d, pd.DataFrame) for d in x.values()]):
            raise TypeError('Values in dict must be pandas.DataFrames.')
    elif isinstance(x, NeuronList):
        if any([not n.has_connectors for n in x]):
            raise ValueError('All neurons must have connector tables as'
                             ' .connectors property.')
    else:
        raise TypeError('Expected Neuronlist or dict of connector tables,'
                        f' got {type(x)}')

    # If single value, turn into list
    if not isinstance(restrict_cn, type(None)):
        restrict_cn = utils.make_iterable(restrict_cn)

    neurons = x.id if isinstance(x, NeuronList) else list(x.keys())

    sim_matrix = pd.DataFrame(np.zeros((len(neurons), len(neurons))),
                              index=neurons,
                              columns=neurons)

    if isinstance(x, NeuronList):
        combinations = [(nA.connectors, nB.connectors, sigma, omega, restrict_cn)
                        for nA in x for nB in x]
        comb_names = [(nA.id, nB.id) for nA in neurons for nB in neurons]
    else:
        combinations = [(nA, nB, sigma, omega, restrict_cn) for nA in x.values() for nB in x.values()]
        comb_names = [(nA, nB) for nA in x.keys() for nB in x.keys()]

    with ThreadPoolExecutor(max_workers=max(1, os.cpu_count())) as e:
        futures = e.map(_unpack_synapse_helper, combinations)

        scores = [n for n in config.tqdm(futures, total=len(combinations),
                                         desc='Processing',
                                         disable=config.pbar_hide,
                                         leave=config.pbar_leave)]

    for c, v in zip(comb_names, scores):
        sim_matrix.loc[c[0], c[1]] = v

    if mu_score:
        sim_matrix = (sim_matrix + sim_matrix.T) / 2

    res = ClustResults(sim_matrix, mat_type='similarity')

    if isinstance(x, NeuronList):
        res.neurons = x  # type: ignore

    return res


def cluster_xyz(x: Union[pd.DataFrame, np.ndarray],
                labels: Optional[List[str]] = None
                ) -> 'ClustResults':
    """ Thin wrapper for ``scipy.scipy.spatial.distance``.

    Takes a list of x,y,z coordinates and calculates EUCLEDIAN distance matrix.

    Parameters
    ----------
    x :             pandas.DataFrame | numpy array (N, 3)
                    If DataFrame, must contain ``x``,``y``,``z`` columns.
    labels :        list of str, optional
                    Labels for each leaf of the dendrogram
                    (e.g. connector ids).

    Returns
    -------
    :class:`navis.ClustResults`
                      Contains distance matrix and methods to generate plots.

    Examples
    --------
    This examples assumes you understand the basics of using navis:

    >>> import matplotlib.pyplot as plt
    >>> n = navis.example_neurons(n=1)
    >>> rs = navis.cluster_xyz(n.connectors,
    ...                        labels=n.connectors.connector_id.values)
    >>> rs.plot_matrix()
    >>> plt.show()

    """
    if isinstance(x, pd.DataFrame):
        # Generate numpy array containing x, y, z coordinates
        try:
            s = x[['x', 'y', 'z']].values
        except BaseException:
            raise ValueError('DataFrame must have x/y/z columns')
    elif isinstance(x, np.ndarray):
        if x.shape[1] != 3:
            raise ValueError('Array must be of shape (N, 3).')
        s = x
    else:
        raise TypeError(f'Expected DataFrame or numpy array, got "{type(x)}".')

    # Calculate euclidean distance matrix
    condensed_dist_mat = scipy.spatial.distance.pdist(s, 'euclidean')
    squared_dist_mat = scipy.spatial.distance.squareform(condensed_dist_mat)

    return ClustResults(squared_dist_mat, labels=labels, mat_type='distance')


class ClustResults:
    """ Class to handle, analyze and plot similarity/distance matrices.

    Contains thin wrappers for ``scipy.cluster``.

    Attributes
    ----------
    dist_mat :  Distance matrix (0=similar, 1=dissimilar)
    sim_mat :   Similarity matrix (0=dissimilar, 1=similar)
    linkage :   Hierarchical clustering. Run :func:`navis.ClustResults.cluster`
                to generate linkage. By default, WARD's algorithm is used.
    leafs :     list of skids

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> # Get a bunch of neurons
    >>> nl = navis.example_neurons()
    >>> # Perform all-by-all nblast
    >>> res = navis.nblast_allbyall(nl)
    >>> # res is a ClustResults object
    >>> res.plot_matrix()
    >>> plt.show()
    >>> # Extract 5 clusters
    >>> res.get_clusters(5, criterion = 'maxclust' )

    """

    _PERM_MAT_TYPES = ['similarity', 'distance']

    neurons: Optional['NeuronList']

    def __init__(self,
                 mat: Union[np.ndarray, pd.DataFrame],
                 labels: Optional[List[str]] = None,
                 mat_type: Union[Literal['distance'],
                                 Literal['similarity']] = 'distance'
                 ):
        """ Initialize class instance.

        Parameters
        ----------
        mat :       numpy.array | pandas.DataFrame
                    Distance or similarity matrix.
        labels :    list, optional
                    Labels for matrix.
        mat_type :  'distance' | 'similarity', default = 'distance'
                    Sets the type of input matrix:
                      - 'similarity' = high values are more similar
                      - 'distance' = low values are more similar

                    The "missing" matrix type will be computed. For clustering,
                    plotting, etc. distance matrices are used.

        """
        if mat_type not in ClustResults._PERM_MAT_TYPES:
            raise ValueError(f'Matrix type "{mat_type}" unkown.')

        if mat_type == 'similarity':
            self.dist_mat = self._invert_mat(mat)
            self.sim_mat = mat
        else:
            self.dist_mat = mat
            self.sim_mat = self._invert_mat(mat)

        self.labels = labels
        self.mat_type = mat_type

        if isinstance(labels, type(None)) and isinstance(mat, pd.DataFrame):
            self.labels = mat.columns.tolist()

    def __getattr__(self, key):
        if key == 'linkage':
            self.cluster()
            return self.linkage
        elif key == 'condensed_dist_mat':
            return scipy.spatial.distance.squareform(self.dist_mat,
                                                     checks=False)
        elif key in ['leafs', 'leaves']:
            return self.get_leafs()
        elif key == 'cophenet':
            return self.calc_cophenet()
        elif key == 'agg_coeff':
            return self.calc_agg_coeff()

    def get_leafs(self, use_labels: bool = False) -> Sequence:
        """ Use to retrieve labels.

        Parameters
        ----------
        use_labels :    bool, optional
                        If True, self.labels will be returned. If False, will
                        use either columns (if matrix is pandas DataFrame)
                        or indices (if matrix is np.ndarray)

        """

        if isinstance(self.dist_mat, pd.DataFrame):
            if use_labels:
                return [self.labels[i] for i in scipy.cluster.hierarchy.leaves_list(self.linkage)]
            else:
                return [self.dist_mat.columns.tolist()[i] for i in scipy.cluster.hierarchy.leaves_list(self.linkage)]
        else:
            return scipy.cluster.hierarchy.leaves_list(self.linkage)

    def calc_cophenet(self) -> float:
        """ Returns Cophenetic Correlation coefficient of your clustering.

        This (very very briefly) compares (correlates) the actual pairwise
        distances of all your samples to those implied by the hierarchical
        clustering. The closer the value is to 1, the better the clustering
        preserves the original distances.

        """

        return scipy.cluster.hierarchy.cophenet(self.linkage,
                                                self.condensed_dist_mat)

    def calc_agg_coeff(self) -> float:
        """ Returns the agglomerative coefficient.

        This measures the clustering structure of the linkage matrix. Because
        it grows with the number of observations, this measure should not be
        used to compare datasets of very different sizes.

        For each observation i, denote by m(i) its dissimilarity to the first
        cluster it is merged with, divided by the dissimilarity of the merger
        in the final step of the algorithm. The agglomerative coefficient is
        the average of all 1 - m(i).

        """

        # Turn into pandas DataFrame for fancy indexing
        Z = pd.DataFrame(self.linkage, columns=['obs1', 'obs2', 'dist', 'n_org'])

        # Get all distances at which an original observation is merged
        all_dist = Z[(Z.obs1.isin(self.leafs)) | (Z.obs2.isin(self.leafs))].dist.values

        # Divide all distances by last merger
        all_dist /= self.linkage[-1][2]

        # Calc final coefficient
        coeff = np.mean(1 - all_dist)

        return coeff

    def _invert_mat(self, sim_mat: pd.DataFrame) -> pd.DataFrame:
        """ Inverts matrix."""
        if isinstance(sim_mat, pd.DataFrame):
            return (sim_mat - sim_mat.max().max()) * -1
        else:
            return (sim_mat - sim_mat.max()) * -1

    def cluster(self, method: str = 'ward') -> None:
        """ Cluster distance matrix.

        This will automatically be called when attribute linkage is requested
        for the first time.

        Parameters
        ----------
        method :    str, optional
                    Clustering method (see scipy.cluster.hierarchy.linkage
                    for reference)

        """

        # Use condensed distance matrix - otherwise clustering thinks we are
        # passing observations instead of final scores
        self.linkage = scipy.cluster.hierarchy.linkage(self.condensed_dist_mat,
                                                       method=method)

        # Save method in case we want to look it up later
        self.cluster_method = method

        logger.info(f'Clustering done using method "{method}"')

    def plot_dendrogram(self,
                        color_threshold: Optional[float] = None,
                        return_dendrogram: bool = False,
                        labels: Optional[List[str]] = None,
                        fig: Optional['matplotlib.figure.Figure'] = None,
                        **kwargs):
        """ Plot dendrogram using matplotlib.

        Parameters
        ----------
        color_threshold :   int | float, optional
                            Coloring threshold for dendrogram.
        return_dendrogram : bool, optional
                            If True, dendrogram object is returned instead of
                            figure.
        labels :            list of str, dict
                            Labels in order of original observation or
                            dictionary with mapping original labels.
        kwargs
                            Passed to ``scipy.cluster.hierarchy.dendrogram()``

        Returns
        -------
        matplotlib.figure
                            If ``return_dendrogram=False`.
        sciyp.cluster.hierarchy.dendrogram
                            If ``return_dendrogram=True``.

        """
        import matplotlib.pyplot as plt

        if isinstance(labels, type(None)):
            labels = self.labels
        elif isinstance(labels, dict):
            labels = [labels[l] for l in self.labels]

        if not fig:
            fig = plt.figure()

        dn_kwargs = {'leaf_rotation': 90,
                     'above_threshold_color': 'k'}
        dn_kwargs.update(kwargs)

        dn = scipy.cluster.hierarchy.dendrogram(self.linkage,
                                                color_threshold=color_threshold,
                                                labels=labels,
                                                **dn_kwargs)
        logger.info(
            'Use matplotlib.pyplot.show() to render dendrogram.')

        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)

        try:
            plt.tight_layout()
        except BaseException:
            pass

        if return_dendrogram:
            return dn
        else:
            return fig

    def plot_matrix2(self, **kwargs):
        """ Plot distance matrix and dendrogram using seaborn. This package
        needs to be installed manually.

        Parameters
        ----------
        kwargs      dict
                    Keyword arguments to be passed to seaborn.clustermap. See
                    http://seaborn.pydata.org/generated/seaborn.clustermap.html


        Returns
        -------
        seaborn.clustermap

        """

        import matplotlib.pyplot as plt

        try:
            import seaborn as sns
        except BaseException:
            raise ImportError('Need seaborn package installed.')

        cg = sns.clustermap(self.dist_mat, row_linkage=self.linkage,
                            col_linkage=self.linkage, **kwargs)

        # Rotate labels
        plt.setp(cg.ax_heatmap.xaxis.get_majorticklabels(), rotation=90)
        plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)

        # Make labels smaller
        plt.setp(cg.ax_heatmap.xaxis.get_majorticklabels(), fontsize=4)
        plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), fontsize=4)

        # Increase padding
        cg.fig.subplots_adjust(right=.8, top=.95, bottom=.2)

        logger.info(
            'Use matplotlib.pyplot.show() to render figure.')

        return cg

    def plot_matrix(self) -> 'matplotlib.figure.Figure':
        """ Plot distance matrix and dendrogram using matplotlib.

        Returns
        -------
        matplotlib figure

        """
        import pylab

        # Compute and plot first dendrogram for all nodes.
        fig = pylab.figure(figsize=(8, 8))
        ax1 = fig.add_axes([0.09, 0.1, 0.2, 0.6])
        Z1 = scipy.cluster.hierarchy.dendrogram(
            self.linkage, orientation='left', labels=self.labels)
        ax1.set_xticks([])
        ax1.set_yticks([])

        # Compute and plot second dendrogram.
        ax2 = fig.add_axes([0.3, 0.71, 0.6, 0.2])
        Z2 = scipy.cluster.hierarchy.dendrogram(
            self.linkage, labels=self.labels)
        ax2.set_xticks([])
        ax2.set_yticks([])

        # Plot distance matrix.
        axmatrix = fig.add_axes([0.3, 0.1, 0.6, 0.6])
        idx1 = Z1['leaves']
        idx2 = Z2['leaves']
        D = self.dist_mat.copy()

        if isinstance(D, pd.DataFrame):
            D = D.values

        D = D[idx1, :]
        D = D[:, idx2]
        im = axmatrix.matshow(
            D, aspect='auto', origin='lower', cmap=pylab.cm.YlGnBu)
        axmatrix.set_xticks([])
        axmatrix.set_yticks([])

        # Plot colorbar.
        axcolor = fig.add_axes([0.91, 0.1, 0.02, 0.6])
        pylab.colorbar(im, cax=axcolor)

        logger.info(
            'Use matplotlib.pyplot.show() to render figure.')

        return fig

    def plot3d(self,
               k: Union[int, float] = 5,
               criterion: str = 'maxclust',
               **kwargs):
        """Plot neuron using :func:`navis.plot.plot3d`. Will only work if
        instance has neurons attached to it.

        Parameters
        ----------
        k :         int | float
        criterion : 'maxclust' | 'distance', optional
                    If ``maxclust``, ``k`` clusters will be formed. If
                    ``distance`` clusters will be created at threshold ``k``.
        **kwargs
                Will be passed to ``navis.plot3d()``. See
                ``help(plot.plot3d)`` for a list of keywords.

        See Also
        --------
        :func:`navis.plot3d`
                    Function called to generate 3d plot.

        """

        if 'neurons' not in self.__dict__:
            logger.error(
                'This works only with cluster results from neurons')
            return None

        cmap = self.get_colormap(k=k, criterion=criterion)

        kwargs.update({'color': cmap})

        return plotting.plot3d(self.neurons, **kwargs)

    def get_colormap(self,
                     k: Union[int, float] = 5,
                     criterion: str = 'maxclust'):
        """Generate colormap based on clustering.

        Parameters
        ----------
        k :         int | float
        criterion : 'maxclust' | 'distance', optional
                    If ``maxclust``, ``k`` clusters will be formed. If
                    ``distance`` clusters will be created at threshold ``k``.

        Returns
        -------
        dict
                    {'skeleton_id': (r,g,b),...}

        """

        cl = self.get_clusters(k, criterion, return_type='indices')

        cl = [[self.dist_mat.index.tolist()[i] for i in l] for l in cl]

        colors = [colorsys.hsv_to_rgb(1 / len(cl) * i, 1, 1)
                  for i in range(len(cl) + 1)]

        return {n: colors[i] for i in range(len(cl)) for n in cl[i]}

    def get_clusters(self,
                     k: Union[int, float],
                     criterion: str = 'maxclust',
                     return_type: str = 'labels') -> List[list]:
        """ Wrapper for ``scipy.cluster.hierarchy.fcluster`` to get clusters.

        Parameters
        ----------
        k :             int | float
        criterion :     'maxclust' | 'distance', optional
                        If ``maxclust``, ``k`` clusters will be formed. If
                        ``distance`` clusters will be created at threshold.
                        ``k``.
        return_type :   'labels' | 'indices' | 'columns' | 'rows'
                        Determines what to construct the clusters of. 'labels'
                        only works if labels are provided. 'indices' refers
                        to index in distance matrix. 'columns'/'rows' works
                        if distance matrix is pandas DataFrame

        Returns
        -------
        list
                    list of clusters ``[[leaf1, leaf5], [leaf2, ...], ...]``

        """

        cl = scipy.cluster.hierarchy.fcluster(
            self.linkage, k, criterion=criterion)

        if not isinstance(self.labels, type(None)) and return_type.lower() == 'labels':
            return [[self.labels[j] for j in range(len(cl)) if cl[j] == i] for i in range(min(cl), max(cl) + 1)]
        elif return_type.lower() == 'rows':
            return [[self.dist_mat.columns.tolist()[j] for j in range(len(cl)) if cl[j] == i] for i in range(min(cl), max(cl) + 1)]
        elif return_type.lower() == 'columns':
            return [[self.dist_mat.index.tolist()[j] for j in range(len(cl)) if cl[j] == i] for i in range(min(cl), max(cl) + 1)]
        else:
            return [[j for j in range(len(cl)) if cl[j] == i] for i in range(min(cl), max(cl) + 1)]

    def to_tree(self):
        """ Turns linkage to ete3 tree.

        See Also
        --------
        http://etetoolkit.org/
            Ete3 homepage

        Returns
        -------
        ete 3 tree

        """
        try:
            import ete3  # ignore type
        except BaseException:
            raise ImportError('Please install ete3 package to use this function.')

        max_dist = self.linkage[-1][2]
        n_original_obs = self.dist_mat.shape[0]

        list_of_childs = {n_original_obs + i: e[:2]
                          for i, e in enumerate(self.linkage)}
        list_of_parents = {int(e[0]): int(n_original_obs + i)
                           for i, e in enumerate(self.linkage)}
        list_of_parents.update(
            {int(e[1]): int(n_original_obs + i) for i, e in enumerate(self.linkage)})

        total_dist = {n_original_obs + i: e[2]
                      for i, e in enumerate(self.linkage)}
        # Process total distance into distances between nodes (root = 0)
        dist_to_parent = {
            n: max_dist - total_dist[list_of_parents[n]] for n in list_of_parents}

        names = {i: n for i, n in enumerate(self.dist_mat.columns.tolist())}

        # Create empty tree
        tree = ete3.Tree()

        # Start with root node
        root = sorted(list(list_of_childs.keys()), reverse=True)[0]
        treenodes = {root: tree.add_child()}

        for k in sorted(list(list_of_childs.keys()), reverse=True):
            e = list_of_childs[k]
            treenodes[e[0]] = treenodes[k].add_child(
                dist=dist_to_parent[e[0]], name=names.get(e[0], None))
            treenodes[e[1]] = treenodes[k].add_child(
                dist=dist_to_parent[e[1]], name=names.get(e[1], None))

        return tree



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import numpy as np
import pandas as pd

from typing import Optional, overload, Union, List
from typing_extensions import Literal

from .. import config, graph, core, utils

# Set up logging
logger = config.logger

__all__ = ['downsample_neuron']


@overload
def downsample_neuron(x: 'core.NeuronObject',
                      downsampling_factor: float,
                      inplace: Literal[True],
                      preserve_nodes: Optional[List[int]] = None
                      ) -> None: ...


@overload
def downsample_neuron(x: 'core.NeuronObject',
                      downsampling_factor: float,
                      inplace: Literal[False],
                      preserve_nodes: Optional[List[int]] = None
                      ) -> 'core.NeuronObject': ...


@overload
def downsample_neuron(x: 'core.NeuronObject',
                      downsampling_factor: float,
                      inplace: bool = False,
                      preserve_nodes: Optional[List[int]] = None
                      ) -> 'core.NeuronObject': ...


def downsample_neuron(x: 'core.NeuronObject',
                      downsampling_factor: Union[int, float],
                      inplace: bool = False,
                      preserve_nodes: Optional[List[int]] = None
                      ) -> Optional['core.NeuronObject']:
    """Downsample neuron(s) by a given factor.

    Preserves root, leafs, branchpoints by default. Preservation of treenodes
    with synapses can be toggled - see ``preserve_nodes`` parameter.

    Notes
    -----
    Use ``downsampling_factor=float('inf')`` to get a neuron consisting only
    of root, branch and end points.

    Parameters
    ----------
    x :                      TreeNeuron | NeuronList
                             Neuron(s) to downsample.
    downsampling_factor :    int | float('inf')
                             Factor by which to reduce the node count.
    preserve_nodes :         str | list, optional
                             Can be either list of node IDs to exclude from
                             downsampling or a string to a DataFrame attached
                             to the neuron (e.g. "connectors"). DataFrame must
                             have `node_id`` column.
    inplace :                bool, optional
                             If True, will modify original neuron. If False, a
                             downsampled copy is returned.

    Returns
    -------
    TreeNeuron/List
                    If ``inplace=False`` (default).

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> n_ds = navis.downsample_neuron(n,
    ...                                downsampling_factor=5,
    ...                                inplace=False)
    >>> n.n_nodes > n_ds.n_nodes
    True

    See Also
    --------
    :func:`navis.resample_neuron`
                             This function resamples a neuron to given
                             resolution. This will change node IDs!

    """
    if isinstance(x, core.NeuronList):
        if not inplace:
            x = x.copy()
        results = [downsample_neuron(x[i],
                                     downsampling_factor=downsampling_factor,
                                     preserve_nodes=preserve_nodes,
                                     inplace=True)
                   for i in config.trange(x.shape[0],
                                          desc='Downsampling',
                                          disable=config.pbar_hide,
                                          leave=config.pbar_leave)]
        if not inplace:
            return core.NeuronList(results)
        return None

    elif isinstance(x, core.TreeNeuron):
        if not inplace:
            x = x.copy()
    else:
        raise TypeError(f'Unable to downsample data of type "{type(x)}"')

    if downsampling_factor <= 1:
        raise ValueError('Downsampling factor must be greater than 1.')

    if not isinstance(preserve_nodes, type(None)):
        if isinstance(preserve_nodes, str):
            table = getattr(x, preserve_nodes)
            if not isinstance(table, pd.DataFrame):
                raise TypeError(f'Expected "{preserve_nodes}" to be a '
                                f'DataFrame - got {type(table)}')
            if 'node_id' not in table.columns:
                raise IndexError(f'DataFrame {preserve_nodes} has no "node_id"'
                                 ' column.')

            preserve_nodes = table['node_id'].values

        if not isinstance(preserve_nodes, (list, set, np.ndarray)):
            raise TypeError('Expected "preserve_nodes" to be list-like, got '
                            f'"{type(preserve_nodes)}"')

    if x.nodes.shape[0] <= 1:
        logger.warning(f'No nodes in neuron {x.id}. Skipping.')
        if not inplace:
            return x
        else:
            return None

    list_of_parents = {n: p for n, p in zip(x.nodes.node_id.values,
                                            x.nodes.parent_id.values)}
    list_of_parents[-1] = -1  # type: ignore  # doesn't know that node_id is int

    if 'type' not in x.nodes:
        graph.classify_nodes(x)

    selection = x.nodes.type != 'slab'

    if utils.is_iterable(preserve_nodes):
        selection = selection | x.nodes.node_id.isin(preserve_nodes)  # type: ignore

    fix_points = x.nodes[selection].node_id.values

    # Add soma node(s)
    if not isinstance(x.soma, type(None)):
        soma = utils.make_iterable(x.soma)
        for s in soma:
            if s not in fix_points:
                fix_points = np.append(fix_points, s)

    # Walk from all fix points to the root - jump N nodes on the way
    new_parents = {}

    for en in fix_points:
        this_node = en

        while True:
            stop = False
            new_p = list_of_parents[this_node]
            if new_p > 0:
                i = 0
                while i < downsampling_factor:
                    if new_p in fix_points or not new_p:
                        new_parents[this_node] = new_p
                        stop = True
                        break
                    new_p = list_of_parents[new_p]
                    i += 1

                if stop is True:
                    break
                else:
                    new_parents[this_node] = new_p
                    this_node = new_p
            else:
                new_parents[this_node] = -1  # type: ignore
                break

    new_nodes = x.nodes[x.nodes.node_id.isin(list(new_parents.keys()))].copy()
    new_nodes.loc[:, 'parent_id'] = [new_parents[tn]
                                     for tn in new_nodes.node_id.values]

    # Assign new parent IDs
    new_nodes.loc[:, 'parent_id'] = new_nodes.parent_id.values.astype(int)

    logger.debug(f'Nodes before/after: {len(x.nodes)}/{len(new_nodes)}')

    x.nodes = new_nodes

    # This is essential -> otherwise e.g. graph.neuron2graph will fail
    x.nodes.reset_index(inplace=True, drop=True)

    x._clear_temp_attr()

    if not inplace:
        return x
    return None



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import collections
import six

import numpy as np
import pandas as pd

from typing import Optional, Any


def make_iterable(x,
                  force_type: Optional[type] = None
                  ) -> np.ndarray:
    """ Helper function. Turns x into a np.ndarray, if it isn't already. For
    dicts, keys will be turned into array.

    Examples
    --------
    >>> from navis.utils import make_iterable
    >>> make_iterable(1)
    array([1])
    >>> make_iterable({'a': 1})
    array(['a'])
    >>> make_iterable([1])
    array([1])
    """
    if not isinstance(x, collections.Iterable) or isinstance(x, six.string_types):
        x = [x]

    if isinstance(x, dict):
        x = list(x)

    if force_type:
        return np.array(x).astype(force_type)
    return np.array(x)


def make_non_iterable(x):
    """ Helper function. Turns x into non-iterable, if it isn't already. Will
    raise error if len(x) > 1.

    Examples
    --------
    >>> from navis.utils import make_non_iterable
    >>> make_non_iterable([1])
    1
    >>> make_non_iterable(1)
    1
    >>> make_non_iterable([1, 2])
    ValueError ...
    """
    if not is_iterable(x):
        return x
    elif len(x) == 1:
        return x[0]
    else:
        raise ValueError('Iterable must not contain more than one entry.')


def is_iterable(x: Any) -> bool:
    """ Helper function. Returns True if x is an iterable but not str.

    Examples
    --------
    >>> from navis.utils import is_iterable
    >>> is_iterable(['a'])
    True
    >>> is_iterable('a')
    False
    >>> is_iterable({'a': 1})
    True
    """
    if isinstance(x, collections.Iterable) and not isinstance(x, (six.string_types, pd.DataFrame)):
        return True
    else:
        return False



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

""" Interface with neuprint. This module is wrapping neuprint-python
(https://github.com/connectome-neuprint/neuprint-python) and adding some
navis-specific functions.
"""

import trimesh

from textwrap import dedent

from concurrent.futures import ThreadPoolExecutor, as_completed

try:
    from neuprint import *
    from neuprint.client import inject_client
except ImportError:
    msg = dedent("""
          neuprint library not found. Please install using pip:

                pip install neuprint-python

          """)
    raise ImportError(msg)
except BaseException:
    raise

import io

from requests.exceptions import HTTPError

import numpy as np
import pandas as pd

from .. import config

from ..core import Volume, TreeNeuron, NeuronList
from ..graph import neuron2KDTree
from ..morpho import heal_fragmented_neuron

logger = config.logger

# Define some integer types
int_types = (int, np.int32, np.int64, np.int, np.int0)


@inject_client
def fetch_roi(roi, *, client=None):
    """Fetch given ROI.

    Parameters
    ----------
    roi :           str
                    Name of an ROI.
    client :        neuprint.Client, optional
                    If ``None`` will try using global client.

    Returns
    -------
    navis.Volume

    """
    if not isinstance(roi, str):
        raise TypeError(f'Expect ROI name as string, got "{type(roi)}"')

    # Fetch data
    data = client.fetch_roi_mesh(roi, export_path=None)

    # Turn into file-like object
    f = io.StringIO(data.decode())

    # Parse with trimesh
    ob = trimesh.load_mesh(f, file_type='obj')

    return Volume.from_object(ob, name=roi)


@inject_client
def fetch_skeletons(x, with_synapses=True, *, heal=False, missing_swc='raise',
                    parallel=True, max_threads=5, client=None):
    """Construct navis.TreeNeuron/List from neuprint neurons.

    Notes
    -----
    Synapses will be attached to the closest node in the skeleton.

    Parameters
    ----------
    x :             str | int | list-like | pandas.DataFrame | SegmentCriteria
                    Body ID(s). Multiple Ids can be provided as list-like or
                    DataFrame with "bodyId" column.
    with_synapses : bool, optional
                    If True will also attach synapses as ``.connectors``.
    heal :          bool, optional
                    If True, will automatically heal fragmented skeletons using
                    neuprint-python's heal function.
    missing_swc :   'raise' | 'warn' | 'skip'
                    What to do if no skeleton is found for a given body ID::

                        "raise" (default) will raise an exception
                        "warn" will throw a warning but continue
                        "skip" will skip without any message

    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    client :        neuprint.Client, optional
                    If ``None`` will try using global client.

    Returns
    -------
    navis.Neuronlist

    """
    if isinstance(x, pd.DataFrame):
        if 'bodyId' in x.columns:
            x = x['bodyId'].values
        else:
            raise ValueError('DataFrame must have "bodyId" column.')

    if isinstance(x, SegmentCriteria):
        query = x
    else:
        query = SegmentCriteria(bodyId=x)

    # Fetch names, etc
    meta, roi_info = fetch_neurons(query, client=client)

    if meta.empty:
        raise ValueError('No neurons matching the given criteria found!')

    # Make sure there is a somaLocation and somaRadius column
    if 'somaLocation' not in meta.columns:
        meta['somaLocation'] = None
        meta['somaRadius'] = None

    nl = []
    with ThreadPoolExecutor(max_workers=1 if not parallel else max_threads) as executor:
        futures = {}
        for r in meta.itertuples():
            f = executor.submit(__fetch_skeleton,
                                r,
                                client=client,
                                with_synapses=with_synapses,
                                missing_swc=missing_swc,
                                heal=heal)
            futures[f] = r.bodyId

        with config.tqdm(desc='Fetching',
                         total=meta.shape[0],
                         leave=config.pbar_leave,
                         disable=meta.shape[0] == 1 or config.pbar_hide) as pbar:
            for f in as_completed(futures):
                bodyId = futures[f]
                pbar.update(1)
                try:
                    nl.append(f.result())
                except Exception as exc:
                    print(f'{bodyId} generated an exception:', exc)

    nl = NeuronList(nl)

    """
    if heal:
        # max_dist of 1000 corresponds to 8um
        heal_fragmented_neuron(nl, max_dist=1000, inplace=True)
    """
    return nl


def __fetch_skeleton(r, client, with_synapses=True, missing_swc='raise', heal=False):
    """Fetch a single skeleton + synapses and turn into CATMAID neuron."""
    # Fetch skeleton SWC
    try:
        data = client.fetch_skeleton(r.bodyId, format='pandas', heal=heal)
    except HTTPError as err:
        if err.response.status_code == 400:
            if missing_swc in ['warn', 'skip']:
                if missing_swc == 'warn':
                    logger.warning(f'No SWC found for {r.bodyId}')
            else:
                raise
        else:
            raise

    # Convert from raw to nanometers
    # TODO!!!

    # Generate neuron
    n = TreeNeuron(data, units='nm')

    # Add some missing meta data
    n.id = r.bodyId

    if hasattr(r, 'instance'):
        n.name = r.instance

    n.n_voxels = r.size
    n.status = r.status

    # Make KDE tree for NN
    if r.somaLocation or with_synapses:
        tree = neuron2KDTree(n, data='nodes')

    # Set soma
    if r.somaLocation:
        d, i = tree.query([r.somaLocation])
        n.soma = int(n.nodes.iloc[i[0]].node_id)
        n.soma_radius = r.somaRadius

    if with_synapses:
        # Fetch synapses
        syn = fetch_synapses(r.bodyId, client=client)

        if not syn.empty:
            # Process synapses
            syn['connector_id'] = syn.index.values
            locs = syn[['x', 'y', 'z']].values
            d, i = tree.query(locs)

            syn['node_id'] = n.nodes.iloc[i].node_id.values
            syn['x'] = locs[:, 0]
            syn['y'] = locs[:, 1]
            syn['z'] = locs[:, 2]

            # Keep only relevant columns
            syn = syn[['connector_id', 'node_id', 'type',
                       'x', 'y', 'z', 'roi', 'confidence']]

            n.connectors = syn

    return n



#    This script is part of navis (http://www.github.com/schlegelp/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.


""" This module contains functions for intersections.
"""

import numpy as np

from typing import Sequence

from ..core import Volume

try:
    from pyoctree import pyoctree
except ImportError:
    pyoctree = None


def ray_in_volume(points: np.ndarray,
                  volume: Volume,
                  multi_ray: bool = False) -> Sequence[bool]:
    """ Uses pyoctree's raycsasting to test if points are within a given
    volume.
    """

    tree = getattr(volume, 'pyoctree', None)

    if not tree:
        # Create octree from scratch
        tree = pyoctree.PyOctree(np.array(volume.vertices, dtype=float, order='C'),
                                 np.array(volume.faces, dtype=np.int32, order='C')
                                 )
        volume.pyoctree = tree  # type: ignore

    # Get min max of volume
    mx = np.array(volume.vertices).max(axis=0)
    mn = np.array(volume.vertices).min(axis=0)

    # Get points outside of bounding box
    bbox_out = (points > mx).any(axis=1) | (points < mn).any(axis=1)
    isin = ~bbox_out
    in_points = points[isin]

    # Perform ray intersection on points inside bounding box
    rayPointList = np.array([[[p[0], mn[1], mn[2]], p] for p in in_points],
                            dtype=np.float32)

    # Get intersections and extract coordinates of intersection
    intersections = [np.array([i.p for i in tree.rayIntersection(ray)]) for ray in rayPointList]

    # In a few odd cases we can get the multiple intersections at the exact
    # same coordinate (something funny with the faces).
    unique_int = [np.unique(np.round(i), axis=0) if np.any(i) else i for i in intersections]

    # Unfortunately rays are bidirectional -> we have to filter intersections
    # to those that occur "above" the point we are querying
    unilat_int = [i[i[:, 2] >= p] if np.any(i) else i for i, p in zip(unique_int, in_points[:, 2])]

    # Count intersections
    int_count = [i.shape[0] for i in unilat_int]

    # Get odd (= in volume) numbers of intersections
    is_odd = np.remainder(int_count, 2) != 0

    # If we want to play it safe, run the above again with two additional rays
    # and decide by majority
    if multi_ray:
        # Run ray from left back
        rayPointList = np.array([[[mn[0], p[1], mn[2]], p] for p in in_points],
                                dtype=np.float32)
        intersections = [np.array([i.p for i in tree.rayIntersection(ray)]) for ray in rayPointList]
        unique_int = [np.unique(i, axis=0) if np.any(i) else i for i in intersections]
        unilat_int = [i[i[:, 0] >= p] if np.any(i) else i for i, p in zip(unique_int, in_points[:, 0])]
        int_count = [i.shape[0] for i in unilat_int]
        is_odd2 = np.remainder(int_count, 2) != 0

        # Run ray from lower left
        rayPointList = np.array([[[mn[0], mn[1], p[2]], p] for p in in_points],
                                dtype=np.float32)
        intersections = [np.array([i.p for i in tree.rayIntersection(ray)]) for ray in rayPointList]
        unique_int = [np.unique(i, axis=0) if np.any(i) else i for i in intersections]
        unilat_int = [i[i[:, 1] >= p] if np.any(i) else i for i, p in zip(unique_int, in_points[:, 1])]
        int_count = [i.shape[0] for i in unilat_int]
        is_odd3 = np.remainder(int_count, 2) != 0

        # Find majority consensus
        is_odd = is_odd.astype(int) + is_odd2.astype(int) + is_odd3.astype(int)
        is_odd = is_odd >= 2

    isin[isin] = is_odd
    return isin
